from contextlib import contextmanager

from dagster import (
    DagsterInstance,
    Definitions,
    _check as check,
    asset,
    define_asset_job,
    observable_source_asset,
    schedule,
    sensor,
)
from dagster._config.field_utils import EnvVar
from dagster._config.pythonic_config import ConfigurableResource
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.test_utils import environ
from dagster_graphql.test.utils import define_out_of_process_context
from pydantic import Field as PyField


class MyResource(ConfigurableResource):
    """My description."""

    a_string: str = "baz"
    an_unset_string: str = "defaulted"


@asset
def my_asset(my_resource: MyResource):
    pass


@observable_source_asset
def my_observable_source_asset(my_resource: MyResource):
    pass


@sensor(asset_selection=AssetSelection.all())
def my_sensor(my_resource: MyResource):
    pass


@sensor(asset_selection=AssetSelection.all())
def my_sensor_two(my_resource: MyResource):
    pass


my_asset_job = define_asset_job(name="my_asset_job", selection=AssetSelection.assets(my_asset))


@schedule(job_name="my_asset_job", cron_schedule="* * * * *")
def my_schedule(my_resource: MyResource):
    pass


class MyInnerResource(ConfigurableResource):
    a_str: str


class MyOuterResource(ConfigurableResource):
    inner: MyInnerResource


class ResourceWithSecrets(ConfigurableResource):
    """A test resource with secret fields."""

    username: str = PyField(description="the username", default="default_user")
    password: str = PyField(
        description="the password",
        default="default_pass",
        json_schema_extra={"dagster__is_secret": True},
    )
    api_key: str = PyField(
        description="the api key",
        default="secret_key_123",
        json_schema_extra={"dagster__is_secret": True},
    )


@asset
def database_asset(my_resource_with_secrets: ResourceWithSecrets):
    """Test asset using database resource."""
    pass


with environ({"MY_STRING": "bar", "MY_OTHER_STRING": "foo"}):
    defs = Definitions(
        assets=[my_asset, my_observable_source_asset, database_asset],
        resources={
            "foo": "a_string",
            "my_resource": MyResource(
                a_string="foo",
            ),
            "my_resource_env_vars": MyResource(a_string=EnvVar("MY_STRING")),
            "my_resource_two_env_vars": MyResource(
                a_string=EnvVar("MY_STRING"), an_unset_string=EnvVar("MY_OTHER_STRING")
            ),
            "my_outer_resource": MyOuterResource(
                inner=MyInnerResource(a_str="wrapped"),
            ),
            "my_resource_with_secrets": ResourceWithSecrets(
                username="admin",
                password="secret123",
            ),
        },
        jobs=[my_asset_job],
        sensors=[my_sensor, my_sensor_two],
        schedules=[my_schedule],
    )


@contextmanager
def define_definitions_test_out_of_process_context(instance):
    check.inst_param(instance, "instance", DagsterInstance)
    with define_out_of_process_context(__file__, "defs", instance) as context:
        yield context



import graphene
from dagster._check import invariant
from dagster._core.utils import is_valid_email


class GrapheneUserDefinitionOwner(graphene.ObjectType):
    class Meta:
        name = "UserDefinitionOwner"

    email = graphene.NonNull(graphene.String)


class GrapheneTeamDefinitionOwner(graphene.ObjectType):
    class Meta:
        name = "TeamDefinitionOwner"

    team = graphene.NonNull(graphene.String)


class GrapheneDefinitionOwner(graphene.Union):
    class Meta:
        types = (
            GrapheneUserDefinitionOwner,
            GrapheneTeamDefinitionOwner,
        )
        name = "DefinitionOwner"


def definition_owner_from_owner_str(
    owner_str: str,
) -> GrapheneUserDefinitionOwner | GrapheneTeamDefinitionOwner:
    if is_valid_email(owner_str):
        return GrapheneUserDefinitionOwner(email=owner_str)
    else:
        invariant(owner_str.startswith("team:"))
        return GrapheneTeamDefinitionOwner(team=owner_str[5:])


# legacy classes for backcompatibility
class GrapheneUserAssetOwner(GrapheneUserDefinitionOwner):
    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        name = "UserAssetOwner"


class GrapheneTeamAssetOwner(GrapheneTeamDefinitionOwner):
    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        name = "TeamAssetOwner"


class GrapheneAssetOwner(graphene.Union):
    class Meta:
        types = (
            GrapheneUserAssetOwner,
            GrapheneTeamAssetOwner,
        )
        name = "AssetOwner"



from datetime import datetime, timezone

from dagster import (
    Definitions,
    LegacyFreshnessPolicy,
    ObserveResult,
    RepositoryDefinition,
    observable_source_asset,
)
from dagster._core.definitions.data_time import DATA_TIME_METADATA_KEY
from dagster._core.definitions.observe import observe
from dagster._core.test_utils import instance_for_test
from dagster_graphql.test.utils import define_out_of_process_context, execute_dagster_graphql

GET_FRESHNESS_INFO = """
    query AssetNodeQuery($assetKeys: [AssetKeyInput!]!) {
        assetNodes(assetKeys: $assetKeys) {
            id
            freshnessInfo {
                currentMinutesLate
                latestMaterializationMinutesLate
            }
            freshnessPolicy {
                cronSchedule
                maximumLagMinutes
            }
        }
    }
"""


@observable_source_asset(legacy_freshness_policy=LegacyFreshnessPolicy(maximum_lag_minutes=30))
def source_asset_with_freshness(context):
    return ObserveResult(metadata={DATA_TIME_METADATA_KEY: float(context.run.tags["data_time"])})


def get_repo() -> RepositoryDefinition:
    return Definitions(assets=[source_asset_with_freshness]).get_repository_def()


def test_source_asset_freshness_policy():
    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as graphql_context:
            result = execute_dagster_graphql(
                graphql_context,
                GET_FRESHNESS_INFO,
                variables={"assetKeys": [{"path": ["source_asset_with_freshness"]}]},
            )
            assert result.data
            assert result.data["assetNodes"][0]["freshnessPolicy"]


def test_source_asset_freshness_info():
    with instance_for_test() as instance:
        with define_out_of_process_context(__file__, "get_repo", instance) as graphql_context:
            result = execute_dagster_graphql(
                graphql_context.reset_for_test(),
                GET_FRESHNESS_INFO,
                variables={"assetKeys": [{"path": ["source_asset_with_freshness"]}]},
            )
            assert result.data
            assert result.data["assetNodes"][0]["freshnessInfo"]["currentMinutesLate"] is None

            observe(
                [source_asset_with_freshness],
                instance=instance,
                tags={"data_time": str(datetime.now(timezone.utc).timestamp() - 1000 * 60 * 35)},
            )

            result = execute_dagster_graphql(
                graphql_context.reset_for_test(),
                GET_FRESHNESS_INFO,
                variables={"assetKeys": [{"path": ["source_asset_with_freshness"]}]},
            )
            assert result.data
            assert result.data["assetNodes"][0]["freshnessInfo"]["currentMinutesLate"] >= 30

            observe(
                [source_asset_with_freshness],
                instance=instance,
                tags={"data_time": str(datetime.now(timezone.utc).timestamp())},
            )

            result = execute_dagster_graphql(
                graphql_context.reset_for_test(),
                GET_FRESHNESS_INFO,
                variables={"assetKeys": [{"path": ["source_asset_with_freshness"]}]},
            )
            assert result.data
            assert result.data["assetNodes"][0]["freshnessInfo"]["currentMinutesLate"] == 0



import graphene
from dagster._core.definitions.partitions.mapping import PartitionMapping


class GraphenePartitionMapping(graphene.ObjectType):
    className = graphene.NonNull(graphene.String)
    description = graphene.NonNull(graphene.String)

    class Meta:
        name = "PartitionMapping"

    def __init__(
        self,
        partition_mapping: PartitionMapping,
    ):
        super().__init__(
            className=type(partition_mapping).__name__,
            description=partition_mapping.description,
        )



import sys

import dagster._check as check
import graphene
from dagster import RunRecord
from dagster._core.definitions.dynamic_partitions_request import (
    AddDynamicPartitionsRequest,
    DeleteDynamicPartitionsRequest,
)
from dagster._core.definitions.run_request import RunRequest
from dagster._core.definitions.schedule_definition import ScheduleExecutionData
from dagster._core.definitions.selector import ScheduleSelector, SensorSelector
from dagster._core.definitions.sensor_definition import SensorExecutionData
from dagster._core.definitions.timestamp import TimestampWithTimezone
from dagster._core.remote_representation.external import CompoundID, RemoteSchedule, RemoteSensor
from dagster._core.scheduler.instigation import (
    DynamicPartitionsRequestResult,
    InstigatorState,
    InstigatorTick,
    InstigatorType,
    ScheduleInstigatorData,
    SensorInstigatorData,
)
from dagster._core.storage.dagster_run import DagsterRun, RunsFilter
from dagster._core.storage.tags import REPOSITORY_LABEL_TAG, TagType, get_tag_type
from dagster._core.utils import is_valid_run_id
from dagster._core.workspace.permissions import Permissions
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster_shared.yaml_utils import dump_run_config_yaml

from dagster_graphql.implementation.fetch_instigators import get_tick_log_events
from dagster_graphql.implementation.fetch_schedules import get_schedule_next_tick
from dagster_graphql.implementation.fetch_sensors import get_sensor_next_tick
from dagster_graphql.implementation.fetch_ticks import get_instigation_ticks
from dagster_graphql.implementation.loader import RepositoryScopedBatchLoader
from dagster_graphql.implementation.utils import (
    UserFacingGraphQLError,
    has_permission_for_definition,
)
from dagster_graphql.schema.entity_key import GrapheneAssetCheckHandle, GrapheneAssetKey
from dagster_graphql.schema.errors import (
    GrapheneError,
    GraphenePythonError,
    GrapheneRepositoryLocationNotFound,
    GrapheneScheduleNotFoundError,
    GrapheneSensorNotFoundError,
)
from dagster_graphql.schema.logs.log_level import GrapheneLogLevel
from dagster_graphql.schema.repository_origin import GrapheneRepositoryOrigin
from dagster_graphql.schema.tags import GraphenePipelineTag
from dagster_graphql.schema.util import ResolveInfo, non_null_list

GrapheneInstigationType = graphene.Enum.from_enum(InstigatorType, "InstigationType")


class GrapheneInstigationStatus(graphene.Enum):
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"

    class Meta:
        name = "InstigationStatus"


class GrapheneInstigationTickStatus(graphene.Enum):
    STARTED = "STARTED"
    SKIPPED = "SKIPPED"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"

    class Meta:
        name = "InstigationTickStatus"


class GrapheneSensorData(graphene.ObjectType):
    lastTickTimestamp = graphene.Float()
    lastRunKey = graphene.String()
    lastCursor = graphene.String()

    class Meta:
        name = "SensorData"

    def __init__(self, instigator_data):
        check.inst_param(instigator_data, "instigator_data", SensorInstigatorData)
        super().__init__(
            lastTickTimestamp=instigator_data.last_tick_timestamp,
            lastRunKey=instigator_data.last_run_key,
            lastCursor=instigator_data.cursor,
        )


class GrapheneScheduleData(graphene.ObjectType):
    cronSchedule = graphene.NonNull(graphene.String)
    startTimestamp = graphene.Float()

    class Meta:
        name = "ScheduleData"

    def __init__(self, instigator_data):
        check.inst_param(instigator_data, "instigator_data", ScheduleInstigatorData)
        super().__init__(
            cronSchedule=str(instigator_data.cron_schedule),
            startTimestamp=instigator_data.start_timestamp,
        )


class GrapheneInstigationTypeSpecificData(graphene.Union):
    class Meta:
        types = (GrapheneSensorData, GrapheneScheduleData)
        name = "InstigationTypeSpecificData"


class GrapheneInstigationEvent(graphene.ObjectType):
    class Meta:
        name = "InstigationEvent"

    message = graphene.NonNull(graphene.String)
    timestamp = graphene.NonNull(graphene.String)
    level = graphene.NonNull(GrapheneLogLevel)


class GrapheneInstigationEventConnection(graphene.ObjectType):
    class Meta:
        name = "InstigationEventConnection"

    events = non_null_list(GrapheneInstigationEvent)
    cursor = graphene.NonNull(graphene.String)
    hasMore = graphene.NonNull(graphene.Boolean)


class GrapheneDynamicPartitionsRequestType(graphene.Enum):
    ADD_PARTITIONS = "ADD_PARTITIONS"
    DELETE_PARTITIONS = "DELETE_PARTITIONS"

    class Meta:
        name = "DynamicPartitionsRequestType"


class DynamicPartitionsRequestMixin:
    # Mixin this class to implement DynamicPartitionsRequest
    #
    # Graphene has some strange properties that make it so that you cannot
    # implement ABCs nor use properties in an overridable way. So the way
    # the mixin works is that the target classes have to have a method
    # get_dynamic_partitions_request()
    partitionKeys = graphene.List(graphene.NonNull(graphene.String))
    partitionsDefName = graphene.NonNull(graphene.String)
    type = graphene.NonNull(GrapheneDynamicPartitionsRequestType)

    class Meta:
        name = "DynamicPartitionRequestMixin"

    def get_dynamic_partitions_request(
        self,
    ) -> AddDynamicPartitionsRequest | DeleteDynamicPartitionsRequest:
        raise NotImplementedError()

    def resolve_partitionKeys(self, _graphene_info: ResolveInfo):
        return self.get_dynamic_partitions_request().partition_keys

    def resolve_partitionsDefName(self, _graphene_info: ResolveInfo):
        return self.get_dynamic_partitions_request().partitions_def_name

    def resolve_type(self, _graphene_info: ResolveInfo):
        return (
            GrapheneDynamicPartitionsRequestType.ADD_PARTITIONS
            if isinstance(self.get_dynamic_partitions_request(), AddDynamicPartitionsRequest)
            else GrapheneDynamicPartitionsRequestType.DELETE_PARTITIONS
        )


class GrapheneDynamicPartitionsRequest(DynamicPartitionsRequestMixin, graphene.ObjectType):
    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        name = "DynamicPartitionRequest"

    def __init__(
        self,
        dynamic_partition_request: AddDynamicPartitionsRequest | DeleteDynamicPartitionsRequest,
    ):
        super().__init__()
        self._dynamic_partitions_request = dynamic_partition_request

    def get_dynamic_partitions_request(
        self,
    ) -> AddDynamicPartitionsRequest | DeleteDynamicPartitionsRequest:
        return self._dynamic_partitions_request


class GrapheneDynamicPartitionsRequestResult(DynamicPartitionsRequestMixin, graphene.ObjectType):
    class Meta:  # pyright: ignore[reportIncompatibleVariableOverride]
        name = "DynamicPartitionsRequestResult"

    skippedPartitionKeys = non_null_list(graphene.String)

    def __init__(self, dynamic_partitions_request_result: DynamicPartitionsRequestResult):
        super().__init__()
        self._dynamic_partitions_request_result = dynamic_partitions_request_result

    def get_dynamic_partitions_request(
        self,
    ) -> AddDynamicPartitionsRequest | DeleteDynamicPartitionsRequest:
        if self._dynamic_partitions_request_result.added_partitions is not None:
            return AddDynamicPartitionsRequest(
                partition_keys=self._dynamic_partitions_request_result.added_partitions,
                partitions_def_name=self._dynamic_partitions_request_result.partitions_def_name,
            )
        elif self._dynamic_partitions_request_result.deleted_partitions is not None:
            return DeleteDynamicPartitionsRequest(
                partition_keys=self._dynamic_partitions_request_result.deleted_partitions,
                partitions_def_name=self._dynamic_partitions_request_result.partitions_def_name,
            )
        else:
            check.failed(
                "Unexpected dynamic_partitions_request_result"
                f" {self._dynamic_partitions_request_result}"
            )

    def resolve_skippedPartitionKeys(self, _graphene_info: ResolveInfo):
        return self._dynamic_partitions_request_result.skipped_partitions


class GrapheneRequestedMaterializationsForAsset(graphene.ObjectType):
    assetKey = graphene.NonNull(GrapheneAssetKey)
    partitionKeys = non_null_list(graphene.String)

    class Meta:
        name = "RequestedMaterializationsForAsset"


class GrapheneInstigationTick(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    tickId = graphene.NonNull(graphene.ID)
    status = graphene.NonNull(GrapheneInstigationTickStatus)
    timestamp = graphene.NonNull(graphene.Float)
    runIds = non_null_list(graphene.String)
    runKeys = non_null_list(graphene.String)
    error = graphene.Field(GraphenePythonError)
    skipReason = graphene.String()
    cursor = graphene.String()
    runs = non_null_list("dagster_graphql.schema.pipelines.pipeline.GrapheneRun")
    originRunIds = non_null_list(graphene.String)
    logKey = graphene.List(graphene.NonNull(graphene.String))
    logEvents = graphene.Field(graphene.NonNull(GrapheneInstigationEventConnection))
    dynamicPartitionsRequestResults = non_null_list(GrapheneDynamicPartitionsRequestResult)
    endTimestamp = graphene.Field(graphene.Float)
    requestedAssetKeys = non_null_list(GrapheneAssetKey)
    requestedAssetMaterializationCount = graphene.NonNull(graphene.Int)
    requestedMaterializationsForAssets = non_null_list(GrapheneRequestedMaterializationsForAsset)
    autoMaterializeAssetEvaluationId = graphene.Field(graphene.ID)
    instigationType = graphene.NonNull(GrapheneInstigationType)

    class Meta:
        name = "InstigationTick"

    def __init__(self, tick: InstigatorTick):
        self._tick = check.inst_param(tick, "tick", InstigatorTick)

        super().__init__(
            status=tick.status.value,
            timestamp=tick.timestamp,
            runIds=tick.run_ids,
            runKeys=tick.run_keys,
            error=GraphenePythonError(tick.error) if tick.error else None,
            instigationType=tick.instigator_type,
            skipReason=tick.skip_reason,
            originRunIds=tick.origin_run_ids,
            cursor=tick.cursor,
            logKey=tick.log_key,
            endTimestamp=tick.end_timestamp,
            autoMaterializeAssetEvaluationId=tick.automation_condition_evaluation_id,
        )

    def resolve_id(self, _):
        return f"{self._tick.instigator_origin_id}:{self._tick.timestamp}"

    def resolve_tickId(self, _: ResolveInfo) -> str:
        return str(self._tick.tick_id)

    async def resolve_runs(self, graphene_info: ResolveInfo):
        from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

        run_ids = self._tick.origin_run_ids or self._tick.run_ids or []

        # filter out backfills
        run_ids = [run_id for run_id in run_ids if is_valid_run_id(run_id)]

        if not run_ids:
            return []

        records_by_id = {
            record.dagster_run.run_id: record
            for record in await RunRecord.gen_many(graphene_info.context, run_ids)
            if record
        }

        return [GrapheneRun(records_by_id[run_id]) for run_id in run_ids if run_id in records_by_id]

    def resolve_logEvents(self, graphene_info: ResolveInfo):
        return get_tick_log_events(graphene_info, self._tick)

    def resolve_dynamicPartitionsRequestResults(self, _):
        return [
            GrapheneDynamicPartitionsRequestResult(request_result)
            for request_result in self._tick.dynamic_partitions_request_results
        ]

    def resolve_requestedAssetKeys(self, _):
        return [
            GrapheneAssetKey(path=asset_key.path) for asset_key in self._tick.requested_asset_keys
        ]

    def resolve_requestedMaterializationsForAssets(self, _):
        return [
            GrapheneRequestedMaterializationsForAsset(
                assetKey=GrapheneAssetKey(path=asset_key.path), partitionKeys=list(partition_keys)
            )
            for asset_key, partition_keys in self._tick.requested_assets_and_partitions.items()
        ]

    def resolve_requestedAssetMaterializationCount(self, _):
        return self._tick.requested_asset_materialization_count


class GrapheneDryRunInstigationTick(graphene.ObjectType):
    timestamp = graphene.Float()
    evaluationResult = graphene.Field(lambda: GrapheneTickEvaluation)

    class Meta:
        name = "DryRunInstigationTick"

    def __init__(
        self,
        selector: ScheduleSelector | SensorSelector,
        timestamp: float | None,
        cursor: str | None = None,
    ):
        self._selector = check.inst_param(selector, "selector", (ScheduleSelector, SensorSelector))
        self._cursor = cursor
        super().__init__(
            timestamp=check.opt_float_param(timestamp, "timestamp"),
        )

    def resolve_evaluationResult(self, graphene_info: ResolveInfo):
        if not graphene_info.context.has_code_location(self._selector.location_name):
            raise UserFacingGraphQLError(
                GrapheneRepositoryLocationNotFound(location_name=self._selector.location_name)
            )

        code_location = graphene_info.context.get_code_location(self._selector.location_name)

        if isinstance(self._selector, SensorSelector):
            sensor = graphene_info.context.get_sensor(self._selector)
            if not sensor:
                raise UserFacingGraphQLError(
                    GrapheneSensorNotFoundError(self._selector.sensor_name)
                )
            sensor_data: SensorExecutionData | SerializableErrorInfo
            try:
                sensor_data = code_location.get_sensor_execution_data(
                    name=self._selector.sensor_name,
                    instance=graphene_info.context.instance,
                    repository_handle=sensor.handle.repository_handle,
                    cursor=self._cursor,
                    last_tick_completion_time=None,
                    last_run_key=None,
                    last_sensor_start_time=None,
                    log_key=None,
                )
            except Exception:
                sensor_data = serializable_error_info_from_exc_info(sys.exc_info())
            return GrapheneTickEvaluation(sensor_data, sensor)
        else:
            schedule = graphene_info.context.get_schedule(self._selector)
            if not schedule:
                raise UserFacingGraphQLError(
                    GrapheneScheduleNotFoundError(self._selector.schedule_name)
                )
            if not self.timestamp:
                raise Exception(
                    "No tick timestamp provided when attempting to dry-run schedule"
                    f" {self._selector.schedule_name}."
                )
            timezone_str = schedule.execution_timezone
            if not timezone_str:
                timezone_str = "UTC"

            next_tick_datetime = next(schedule.execution_time_iterator(self.timestamp))
            schedule_data: ScheduleExecutionData | SerializableErrorInfo
            try:
                schedule_data = code_location.get_schedule_execution_data(
                    instance=graphene_info.context.instance,
                    repository_handle=schedule.handle.repository_handle,
                    schedule_name=schedule.name,
                    scheduled_execution_time=TimestampWithTimezone(
                        next_tick_datetime.timestamp(),
                        timezone_str,
                    ),
                    log_key=None,
                )
            except Exception:
                schedule_data = serializable_error_info_from_exc_info(sys.exc_info())

            return GrapheneTickEvaluation(schedule_data, schedule)


class GrapheneTickEvaluation(graphene.ObjectType):
    dynamicPartitionsRequests = graphene.List(graphene.NonNull(GrapheneDynamicPartitionsRequest))
    runRequests = graphene.List(lambda: graphene.NonNull(GrapheneRunRequest))
    skipReason = graphene.String()
    error = graphene.Field(GraphenePythonError)
    cursor = graphene.String()

    _execution_data: ScheduleExecutionData | SensorExecutionData | SerializableErrorInfo

    class Meta:
        name = "TickEvaluation"

    def __init__(
        self,
        execution_data: ScheduleExecutionData | SensorExecutionData | SerializableErrorInfo,
        instigator: RemoteSensor | RemoteSchedule,
    ):
        check.inst_param(
            execution_data,
            "execution_data",
            (ScheduleExecutionData, SensorExecutionData, SerializableErrorInfo),
        )
        error = (
            GraphenePythonError(execution_data)
            if isinstance(execution_data, SerializableErrorInfo)
            else None
        )
        skip_reason = (
            execution_data.skip_message
            if not isinstance(execution_data, SerializableErrorInfo)
            else None
        )
        self._execution_data = execution_data
        self._run_requests = (
            execution_data.run_requests
            if not isinstance(execution_data, SerializableErrorInfo)
            else None
        )

        self._dynamic_partitions_requests = (
            execution_data.dynamic_partitions_requests
            if isinstance(execution_data, SensorExecutionData)
            else None
        )

        cursor = execution_data.cursor if isinstance(execution_data, SensorExecutionData) else None
        self._instigator = instigator
        super().__init__(
            skipReason=skip_reason,
            error=error,
            cursor=cursor,
        )

    def resolve_runRequests(self, _graphene_info: ResolveInfo):
        from dagster._daemon.sensor import fetch_existing_runs

        if not self._run_requests:
            return self._run_requests

        # filter out run requests that already have runs matching their run_keys
        if isinstance(self._instigator, RemoteSensor):
            existing_runs = fetch_existing_runs(
                _graphene_info.context.instance,
                self._instigator,
                self._run_requests,
            )
            valid_run_requests = [
                run_request
                for run_request in self._run_requests
                if run_request.run_key not in existing_runs
            ]
        else:
            valid_run_requests = self._run_requests

        return [GrapheneRunRequest(run_request) for run_request in valid_run_requests]

    def resolve_dynamicPartitionsRequests(self, _graphene_info: ResolveInfo):
        if not self._dynamic_partitions_requests:
            return self._dynamic_partitions_requests

        return [
            GrapheneDynamicPartitionsRequest(request)
            for request in self._dynamic_partitions_requests
        ]


class GrapheneRunRequest(graphene.ObjectType):
    runKey = graphene.String()
    tags = non_null_list(GraphenePipelineTag)
    runConfigYaml = graphene.NonNull(graphene.String)
    assetSelection = graphene.List(graphene.NonNull(GrapheneAssetKey))
    assetChecks = graphene.List(graphene.NonNull(GrapheneAssetCheckHandle))
    jobName = graphene.String()

    _run_request: RunRequest

    class Meta:
        name = "RunRequest"

    def __init__(self, run_request):
        super().__init__(runKey=run_request.run_key)
        self._run_request = check.inst_param(run_request, "run_request", RunRequest)

    def resolve_tags(self, _graphene_info: ResolveInfo):
        return [
            GraphenePipelineTag(key=key, value=value)
            for key, value in self._run_request.tags.items()
            if get_tag_type(key) != TagType.HIDDEN
        ]

    def resolve_runConfigYaml(self, _graphene_info: ResolveInfo):
        return dump_run_config_yaml(self._run_request.run_config)

    def resolve_assetSelection(self, _graphene_info: ResolveInfo):
        if self._run_request.asset_selection:
            return [
                GrapheneAssetKey(path=asset_key.path)
                for asset_key in self._run_request.asset_selection
            ]
        return None

    def resolve_jobName(self, _graphene_info: ResolveInfo):
        return self._run_request.job_name

    def resolve_assetChecks(self, _graphene_info: ResolveInfo):
        if self._run_request.asset_check_keys:
            return [
                GrapheneAssetCheckHandle(check_key)
                for check_key in self._run_request.asset_check_keys
            ]
        return None


class GrapheneDryRunInstigationTicks(graphene.ObjectType):
    results = non_null_list(GrapheneDryRunInstigationTick)
    cursor = graphene.NonNull(graphene.Float)

    class Meta:
        name = "DryRunInstigationTicks"


class GrapheneInstigationState(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    selectorId = graphene.NonNull(graphene.String)
    name = graphene.NonNull(graphene.String)
    instigationType = graphene.NonNull(GrapheneInstigationType)
    status = graphene.NonNull(GrapheneInstigationStatus)
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)
    repositoryOrigin = graphene.NonNull(GrapheneRepositoryOrigin)
    typeSpecificData = graphene.Field(GrapheneInstigationTypeSpecificData)
    runs = graphene.Field(
        non_null_list("dagster_graphql.schema.pipelines.pipeline.GrapheneRun"),
        limit=graphene.Int(),
    )
    runsCount = graphene.NonNull(graphene.Int)
    tick = graphene.Field(
        graphene.NonNull(GrapheneInstigationTick),
        tickId=graphene.NonNull(graphene.ID),
    )
    ticks = graphene.Field(
        non_null_list(GrapheneInstigationTick),
        dayRange=graphene.Int(),
        dayOffset=graphene.Int(),
        limit=graphene.Int(),
        cursor=graphene.String(),
        statuses=graphene.List(graphene.NonNull(GrapheneInstigationTickStatus)),
        beforeTimestamp=graphene.Float(),
        afterTimestamp=graphene.Float(),
    )
    nextTick = graphene.Field(GrapheneDryRunInstigationTick)
    runningCount = graphene.NonNull(graphene.Int)  # remove with cron scheduler

    hasStartPermission = graphene.NonNull(graphene.Boolean)
    hasStopPermission = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "InstigationState"

    def __init__(
        self,
        instigator_state: InstigatorState,
        batch_loader=None,
    ):
        self._instigator_state = check.inst_param(
            instigator_state, "instigator_state", InstigatorState
        )

        # optional batch loader, provided by a parent GrapheneRepository object that instantiates
        # multiple schedules/sensors
        self._batch_loader = check.opt_inst_param(
            batch_loader, "batch_loader", RepositoryScopedBatchLoader
        )
        cid = CompoundID(
            remote_origin_id=instigator_state.instigator_origin_id,
            selector_id=instigator_state.selector_id,
        )
        super().__init__(
            id=cid.to_string(),
            selectorId=instigator_state.selector_id,
            name=instigator_state.name,
            instigationType=instigator_state.instigator_type.value,
            status=(
                GrapheneInstigationStatus.RUNNING
                if instigator_state.is_running
                else GrapheneInstigationStatus.STOPPED
            ),
        )

    def resolve_repositoryOrigin(self, _graphene_info: ResolveInfo):
        origin = self._instigator_state.origin.repository_origin
        return GrapheneRepositoryOrigin(origin)

    def resolve_repositoryName(self, _graphene_info: ResolveInfo):
        return self._instigator_state.repository_selector.repository_name

    def resolve_repositoryLocationName(self, _graphene_info: ResolveInfo):
        return self._instigator_state.repository_selector.location_name

    def _has_permission(self, graphene_info: ResolveInfo, permission: Permissions) -> bool:
        is_schedule = self._instigator_state.instigator_type == InstigatorType.SCHEDULE
        if not self._batch_loader:
            if is_schedule:
                selector = ScheduleSelector.from_instigator_selector(
                    self._instigator_state.selector
                )
            else:
                selector = SensorSelector.from_instigator_selector(self._instigator_state.selector)

            return graphene_info.context.has_permission_for_selector(permission, selector)

        # we have the repository in scope, we should just check off of the definition instead of accessing
        # the selector-based permission check using a cached call
        repository = self._batch_loader.repository
        has_definition = (
            repository.has_schedule(self._instigator_state.name)
            if is_schedule
            else repository.has_sensor(self._instigator_state.name)
        )
        if not has_definition:
            return graphene_info.context.has_permission_for_location(
                permission,
                self._instigator_state.repository_selector.location_name,
            )

        definition = (
            repository.get_schedule(self._instigator_state.name)
            if is_schedule
            else repository.get_sensor(self._instigator_state.name)
        )
        return has_permission_for_definition(graphene_info, permission, definition)

    def resolve_hasStartPermission(self, graphene_info: ResolveInfo):
        permission = (
            Permissions.START_SCHEDULE
            if self._instigator_state.instigator_type == InstigatorType.SCHEDULE
            else Permissions.EDIT_SENSOR
        )
        return self._has_permission(graphene_info, permission)

    def resolve_hasStopPermission(self, graphene_info: ResolveInfo):
        permission = (
            Permissions.STOP_RUNNING_SCHEDULE
            if self._instigator_state.instigator_type == InstigatorType.SCHEDULE
            else Permissions.EDIT_SENSOR
        )
        return self._has_permission(graphene_info, permission)

    def resolve_typeSpecificData(self, _graphene_info: ResolveInfo):
        if not self._instigator_state.instigator_data:
            return None

        if self._instigator_state.instigator_type == InstigatorType.SENSOR:
            return GrapheneSensorData(self._instigator_state.instigator_data)

        if self._instigator_state.instigator_type == InstigatorType.SCHEDULE:
            return GrapheneScheduleData(self._instigator_state.instigator_data)

        return None

    def resolve_runs(self, graphene_info: ResolveInfo, limit: int | None = None):
        from dagster_graphql.schema.pipelines.pipeline import GrapheneRun

        repository_label = self._instigator_state.origin.repository_origin.get_label()
        if self._instigator_state.instigator_type == InstigatorType.SENSOR:
            filters = RunsFilter(
                tags={
                    **DagsterRun.tags_for_sensor(self._instigator_state),
                    REPOSITORY_LABEL_TAG: repository_label,
                }
            )
        else:
            filters = RunsFilter(
                tags={
                    **DagsterRun.tags_for_schedule(self._instigator_state),
                    REPOSITORY_LABEL_TAG: repository_label,
                }
            )
        return [
            GrapheneRun(record)
            for record in graphene_info.context.instance.get_run_records(
                filters=filters,
                limit=limit,
            )
        ]

    def resolve_runsCount(self, graphene_info: ResolveInfo):
        if self._instigator_state.instigator_type == InstigatorType.SENSOR:
            filters = RunsFilter.for_sensor(self._instigator_state)
        else:
            filters = RunsFilter.for_schedule(self._instigator_state)
        return graphene_info.context.instance.get_runs_count(filters=filters)

    def resolve_tick(
        self,
        graphene_info: ResolveInfo,
        tickId: str,
    ) -> GrapheneInstigationTick:
        schedule_storage = check.not_none(graphene_info.context.instance.schedule_storage)
        tick = schedule_storage.get_tick(int(tickId))

        return GrapheneInstigationTick(tick)

    def resolve_ticks(
        self,
        graphene_info,
        dayRange=None,
        dayOffset=None,
        limit=None,
        cursor=None,
        statuses=None,
        beforeTimestamp=None,
        afterTimestamp=None,
    ):
        return get_instigation_ticks(
            graphene_info=graphene_info,
            instigator_type=self._instigator_state.instigator_type,
            instigator_origin_id=self._instigator_state.instigator_origin_id,
            selector_id=self._instigator_state.selector_id,
            batch_loader=self._batch_loader,
            dayRange=dayRange,
            dayOffset=dayOffset,
            limit=limit,
            cursor=cursor,
            status_strings=statuses,
            before=beforeTimestamp,
            after=afterTimestamp,
        )

    def resolve_nextTick(self, graphene_info: ResolveInfo):
        # sensor
        if self._instigator_state.instigator_type == InstigatorType.SENSOR:
            return get_sensor_next_tick(graphene_info, self._instigator_state)
        else:
            return get_schedule_next_tick(graphene_info, self._instigator_state)

    def resolve_runningCount(self, _graphene_info: ResolveInfo):
        return 1 if self._instigator_state.is_running else 0


class GrapheneInstigationStates(graphene.ObjectType):
    results = non_null_list(GrapheneInstigationState)

    class Meta:
        name = "InstigationStates"


class GrapheneInstigationStateNotFoundError(graphene.ObjectType):
    class Meta:
        interfaces = (GrapheneError,)
        name = "InstigationStateNotFoundError"

    name = graphene.NonNull(graphene.String)

    def __init__(self, target):
        super().__init__()
        self.name = check.str_param(target, "target")
        self.message = f"Could not find instigation state for `{target}`"


class GrapheneInstigationStateOrError(graphene.Union):
    class Meta:
        name = "InstigationStateOrError"
        types = (
            GrapheneInstigationState,
            GrapheneInstigationStateNotFoundError,
            GraphenePythonError,
        )


class GrapheneInstigationStatesOrError(graphene.Union):
    class Meta:
        types = (GrapheneInstigationStates, GraphenePythonError)
        name = "InstigationStatesOrError"


types = [
    GrapheneDryRunInstigationTick,
    GrapheneDryRunInstigationTicks,
    GrapheneInstigationTypeSpecificData,
    GrapheneInstigationState,
    GrapheneInstigationStateNotFoundError,
    GrapheneInstigationStateOrError,
    GrapheneInstigationStates,
    GrapheneInstigationStatesOrError,
    GrapheneInstigationTick,
    GrapheneScheduleData,
    GrapheneSensorData,
]



"""Tests for secret field functionality in GraphQL config types."""

from dagster._core.test_utils import instance_for_test
from dagster_graphql.test.utils import execute_dagster_graphql

from dagster_graphql_tests.graphql.repo_definitions import (
    define_definitions_test_out_of_process_context,
)

# Query to fetch config type information for a resource
CONFIG_TYPE_QUERY = """
query ResourceConfigQuery($selector: PipelineSelector!) {
    pipelineOrError(params: $selector) {
        __typename
        ... on Pipeline {
            name
            modes {
                name
                resources {
                    name
                    configField {
                        configType {
                            key
                            ... on CompositeConfigType {
                                fields {
                                    name
                                    description
                                    isRequired
                                    isSecret
                                    defaultValueAsJson
                                    configType {
                                        key
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}
"""


def test_secret_field_in_resource_config():
    """Test that is_secret field is properly exposed in GraphQL for resource config."""
    with instance_for_test() as instance:
        with define_definitions_test_out_of_process_context(instance) as context:
            result = execute_dagster_graphql(
                context,
                CONFIG_TYPE_QUERY,
                variables={
                    "selector": {
                        "pipelineName": "__ASSET_JOB",
                        "repositoryName": "__repository__",
                        "repositoryLocationName": context.code_location_names[0],
                    }
                },
            )

            assert not result.errors
            assert result.data

            # Get the pipeline data
            pipeline_data = result.data["pipelineOrError"]
            assert pipeline_data["__typename"] == "Pipeline"

            # Find the database resource
            resources = pipeline_data["modes"][0]["resources"]
            secrets_resource = next(
                (r for r in resources if r["name"] == "my_resource_with_secrets"), None
            )

            # If database resource not found, skip test (may not be in all test variants)
            if not secrets_resource:
                return

            # Get the config fields
            config_fields = secrets_resource["configField"]["configType"]["fields"]

            # Find each field and check is_secret and defaultValueAsJson
            field_dict = {field["name"]: field for field in config_fields}

            assert "username" in field_dict
            assert field_dict["username"]["isSecret"] is False
            # Non-secret field should have actual default value
            assert field_dict["username"]["defaultValueAsJson"] == '"default_user"'

            # Secret fields should be masked
            assert "password" in field_dict
            assert field_dict["password"]["isSecret"] is True
            assert field_dict["password"]["defaultValueAsJson"] == '"********"'

            assert "api_key" in field_dict
            assert field_dict["api_key"]["isSecret"] is True
            assert field_dict["api_key"]["defaultValueAsJson"] == '"********"'



import asyncio

import graphene

_STATE = {}


class GrapheneTestFields(graphene.ObjectType):
    class Meta:
        name = "TestFields"

    alwaysException = graphene.String()
    asyncString = graphene.String()

    def resolve_alwaysException(self, _):
        raise Exception("as advertised")

    async def resolve_asyncString(self, _):
        msg = "slept"
        if _STATE.get("sleeping"):
            msg += " concurrently"
        else:
            _STATE["sleeping"] = True

        await asyncio.sleep(0)
        _STATE["sleeping"] = False
        return msg



import functools
from unittest.mock import Mock

import dagster._check as check
import pytest
from dagster._core.workspace.permissions import (
    EDITOR_PERMISSIONS,
    VIEWER_PERMISSIONS,
    Permissions,
    get_location_scoped_user_permissions,
)
from dagster_graphql.implementation.utils import (
    UserFacingGraphQLError,
    assert_permission,
    assert_permission_for_location,
    capture_error,
    check_permission,
    require_permission_check,
)
from dagster_graphql.schema.util import ResolveInfo
from dagster_graphql.test.utils import execute_dagster_graphql

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    NonLaunchableGraphQLContextTestMatrix,
)

PERMISSIONS_QUERY = """
    query PermissionsQuery {
      permissions {
        permission
        value
        disabledReason
      }
    }
"""

WORKSPACE_PERMISSIONS_QUERY = """
query WorkspacePermissionsQuery {
  # faster loading option due to avoiding full workspace load
  locationStatusesOrError {
    __typename
    ... on WorkspaceLocationStatusEntries {
      entries {
        permissions {
          permission
          value
          disabledReason
        }
      }
    }
  }
  workspaceOrError {
    ... on Workspace {
      locationEntries {
        permissions {
          permission
          value
          disabledReason
        }
      }
    }
  }
}
"""


class FakeMutation:
    @check_permission("fake_permission")
    def mutate(self, graphene_info: ResolveInfo, **_kwargs):
        pass


class FakeMutationAsync:
    @check_permission("fake_permission")
    async def mutate(self, graphene_info: ResolveInfo, **_kwargs):
        pass


class FakeOtherPermissionMutation:
    @check_permission("fake_other_permission")
    def mutate(self, graphene_info: ResolveInfo, **_kwargs):
        pass


class FakeOtherPermissionMutationAsync:
    @check_permission("fake_other_permission")
    async def mutate(self, graphene_info: ResolveInfo, **_kwargs):
        pass


class FakeMissingPermissionMutation:
    @check_permission("fake_missing_permission")
    def mutate(self, graphene_info: ResolveInfo, **_kwargs):
        pass


class FakeEnumPermissionMutation:
    @check_permission(Permissions.LAUNCH_PIPELINE_EXECUTION)
    def mutate(self, graphene_info: ResolveInfo, **_kwargs):
        pass


class FakeOtherEnumPermissionMutation:
    @check_permission(Permissions.LAUNCH_PIPELINE_REEXECUTION)
    def mutate(self, graphene_info: ResolveInfo, **_kwargs):
        pass


class FakeMissingEnumPermissionMutation:
    @check_permission(Permissions.LAUNCH_PARTITION_BACKFILL)
    def mutate(self, graphene_info: ResolveInfo, **_kwargs):
        pass


class EndpointWithRequiredPermissionCheck:
    @require_permission_check(Permissions.LAUNCH_PIPELINE_EXECUTION)
    def mutate(self, graphene_info, **_kwargs):
        assert_permission(graphene_info, Permissions.LAUNCH_PIPELINE_EXECUTION)


def decorator_with_attribute(fn):
    @functools.wraps(fn)
    def _fn(*args, **kwargs):
        return fn(*args, **kwargs)

    fn.my_attribute = True

    return _fn


def test_decorators_wrap_attributes():
    class EndpointWithAttributeSet:
        @capture_error
        @check_permission(Permissions.LAUNCH_PARTITION_BACKFILL)
        @decorator_with_attribute
        def mutate(self, graphene_info, **_kwargs):
            pass

    assert EndpointWithAttributeSet.mutate.__wrapped__.__wrapped__.__wrapped__.my_attribute is True  # pyright: ignore[reportFunctionMemberAccess]


class EndpointMissingRequiredPermissionCheck:
    @require_permission_check(Permissions.LAUNCH_PARTITION_BACKFILL)
    def mutate(self, graphene_info, **_kwargs):
        pass


class EndpointMissingRequiredPermissionCheckAsync:
    @require_permission_check(Permissions.LAUNCH_PARTITION_BACKFILL)
    async def mutate(self, graphene_info, **_kwargs):
        pass


class FakeEnumLocationPermissionMutation:
    @require_permission_check(Permissions.LAUNCH_PIPELINE_EXECUTION)
    def mutate(self, graphene_info, **kwargs):
        location_name = kwargs["locationName"]
        assert_permission_for_location(
            graphene_info, Permissions.LAUNCH_PIPELINE_EXECUTION, location_name
        )


class FakeEnumLocationPermissionMutationAsync:
    @require_permission_check(Permissions.LAUNCH_PIPELINE_EXECUTION)
    async def mutate(self, graphene_info, **kwargs):
        location_name = kwargs["locationName"]
        assert_permission_for_location(
            graphene_info, Permissions.LAUNCH_PIPELINE_EXECUTION, location_name
        )


@pytest.fixture(name="fake_graphene_info")
def fake_graphene_info_fixture():
    context = Mock()
    _did_check = {}

    def fake_has_permission(permission):
        permission_map = {
            "fake_permission": True,
            "fake_other_permission": False,
            Permissions.LAUNCH_PIPELINE_EXECUTION: True,
            Permissions.LAUNCH_PIPELINE_REEXECUTION: False,
        }
        check.invariant(permission in permission_map, "Permission does not exist in map")
        _did_check[permission] = True
        return permission_map[permission]

    context.has_permission = Mock(side_effect=fake_has_permission)

    def fake_has_permission_for_location(permission, location_name):
        _did_check[permission] = True
        return location_name == "has_location_permission"

    def fake_was_permission_checked(_permission):
        return _did_check.get(_permission) or False

    context.has_permission_for_location = Mock(side_effect=fake_has_permission_for_location)

    context.was_permission_checked = Mock(side_effect=fake_was_permission_checked)

    graphene_info = Mock()
    graphene_info.context = context

    yield graphene_info


def test_check_location_mutation(fake_graphene_info):
    # Fails per-location check
    mutation = FakeEnumLocationPermissionMutation()
    mutation.mutate(fake_graphene_info, locationName="has_location_permission")

    with pytest.raises(UserFacingGraphQLError, match="GrapheneUnauthorizedError"):
        mutation.mutate(fake_graphene_info, locationName="no_location_permission")


@pytest.mark.asyncio
async def test_check_location_mutation_async(fake_graphene_info):
    # Fails per-location check
    mutation = FakeEnumLocationPermissionMutationAsync()
    await mutation.mutate(fake_graphene_info, locationName="has_location_permission")

    with pytest.raises(UserFacingGraphQLError, match="GrapheneUnauthorizedError"):
        await mutation.mutate(fake_graphene_info, locationName="no_location_permission")


def test_require_permission_check_succeeds(fake_graphene_info):
    # Fails per-location check
    mutation = EndpointWithRequiredPermissionCheck()
    mutation.mutate(fake_graphene_info)


def test_require_permission_check_missing(fake_graphene_info):
    # Fails per-location check
    mutation = EndpointMissingRequiredPermissionCheck()
    with pytest.raises(
        Exception, match="Permission launch_partition_backfill was never checked during the request"
    ):
        mutation.mutate(fake_graphene_info)


@pytest.mark.asyncio
async def test_require_permission_check_missing_async(fake_graphene_info):
    # Fails per-location check
    mutation = EndpointMissingRequiredPermissionCheckAsync()
    with pytest.raises(
        Exception, match="Permission launch_partition_backfill was never checked during the request"
    ):
        await mutation.mutate(fake_graphene_info)


@pytest.mark.parametrize("mutation", [FakeMutation(), FakeEnumPermissionMutation()])
def test_check_permission_has_permission(fake_graphene_info, mutation):
    mutation.mutate(fake_graphene_info)


@pytest.mark.asyncio
async def test_check_permission_has_permission_async(fake_graphene_info):
    await FakeMutationAsync().mutate(fake_graphene_info)


@pytest.mark.parametrize(
    "mutation", [FakeOtherPermissionMutation(), FakeOtherEnumPermissionMutation()]
)
def test_check_permission_does_not_have_permission(fake_graphene_info, mutation):
    with pytest.raises(UserFacingGraphQLError, match="GrapheneUnauthorizedError"):
        mutation.mutate(fake_graphene_info)


@pytest.mark.asyncio
async def test_check_permission_does_not_have_permission_async(fake_graphene_info):
    with pytest.raises(UserFacingGraphQLError, match="GrapheneUnauthorizedError"):
        await FakeOtherPermissionMutationAsync().mutate(fake_graphene_info)


@pytest.mark.parametrize(
    "mutation", [FakeMissingPermissionMutation(), FakeMissingEnumPermissionMutation()]
)
def test_check_permission_permission_does_not_exist(fake_graphene_info, mutation):
    with pytest.raises(check.CheckError):
        mutation.mutate(fake_graphene_info)


@pytest.mark.parametrize("permission", ["fake_permission", Permissions.LAUNCH_PIPELINE_EXECUTION])
def test_assert_permission_has_permission(fake_graphene_info, permission):
    assert_permission(fake_graphene_info, permission)


@pytest.mark.parametrize(
    "permission", ["fake_other_permission", Permissions.LAUNCH_PIPELINE_REEXECUTION]
)
def test_assert_permission_does_not_have_permission(fake_graphene_info, permission):
    with pytest.raises(UserFacingGraphQLError, match="GrapheneUnauthorizedError"):
        assert_permission(fake_graphene_info, permission)


@pytest.mark.parametrize(
    "permission", ["fake_missing_permission", Permissions.LAUNCH_PARTITION_BACKFILL]
)
def test_assert_permission_permission_does_not_exist(fake_graphene_info, permission):
    with pytest.raises(check.CheckError):
        assert_permission(fake_graphene_info, permission)


class TestPermissionsQuery(NonLaunchableGraphQLContextTestMatrix):
    def test_permissions_query(self, graphql_context):
        result = execute_dagster_graphql(graphql_context, PERMISSIONS_QUERY)
        assert result.data

        assert result.data["permissions"]

        permissions_map = {
            permission["permission"]: permission["value"]
            for permission in result.data["permissions"]
        }

        if graphql_context.read_only:
            assert permissions_map == VIEWER_PERMISSIONS
        else:
            assert permissions_map == EDITOR_PERMISSIONS

        for permission in result.data["permissions"]:
            if not permission["value"]:
                assert permission["disabledReason"] == "Disabled by your administrator"
            else:
                assert not permission.get("disabledReason")

    def test_truthy_permission_result(self, graphql_context):
        permissions = graphql_context.permissions
        permission_result = next(iter(permissions.values()))

        with pytest.raises(
            Exception,
            match=(
                "Don't check a PermissionResult for truthiness - check the `enabled` property"
                " instead"
            ),
        ):
            if permission_result:
                pass

        permission_result.enabled  # noqa: B018


class TestWorkspacePermissionsQuery(NonLaunchableGraphQLContextTestMatrix):
    def test_workspace_permissions_query(self, graphql_context):
        result = execute_dagster_graphql(graphql_context, WORKSPACE_PERMISSIONS_QUERY)
        assert result.data

        assert result.data["workspaceOrError"]["locationEntries"]
        assert result.data["locationStatusesOrError"]["entries"]

        # ensure both old and new ways to fetch work and return same data
        assert (
            result.data["locationStatusesOrError"]["entries"]
            == result.data["workspaceOrError"]["locationEntries"]
        )

        for location in result.data["locationStatusesOrError"]["entries"]:
            permissions_map = {
                permission["permission"]: permission["value"]
                for permission in location["permissions"]
            }

            expected_permissions_map = {
                key: perm.enabled
                for key, perm in get_location_scoped_user_permissions(
                    graphql_context.read_only
                ).items()
            }
            assert permissions_map == expected_permissions_map



from enum import Enum
from typing import Any, NamedTuple


class DagsterGraphQLClientError(Exception):
    def __init__(self, *args, body=None):
        super().__init__(*args)
        self.body = body


class ReloadRepositoryLocationStatus(Enum):
    """This enum describes the status of a GraphQL mutation to reload a Dagster repository location.

    Args:
        Enum (str): can be either `ReloadRepositoryLocationStatus.SUCCESS`
            or `ReloadRepositoryLocationStatus.FAILURE`.
    """

    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


class ShutdownRepositoryLocationStatus(Enum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


class ReloadRepositoryLocationInfo(NamedTuple):
    """This class gives information about the result of reloading
    a Dagster repository location with a GraphQL mutation.

    Args:
        status (ReloadRepositoryLocationStatus): The status of the reload repository location mutation
        failure_type: (Optional[str], optional): the failure type if `status == ReloadRepositoryLocationStatus.FAILURE`.
          Can be one of `ReloadNotSupported`, `RepositoryLocationNotFound`, or `RepositoryLocationLoadFailure`. Defaults to None.
        message (Optional[str], optional): the failure message/reason if
          `status == ReloadRepositoryLocationStatus.FAILURE`. Defaults to None.
    """

    status: ReloadRepositoryLocationStatus
    failure_type: str | None = None
    message: str | None = None


class ShutdownRepositoryLocationInfo(NamedTuple):
    """This class gives information about the result of shutting down the server for
    a Dagster repository location using a GraphQL mutation.

    Args:
        status (ShutdownRepositoryLocationStatus) Whether the shutdown succeeded or failed.
        message (Optional[str], optional): the failure message/reason if
          `status == ShutdownRepositoryLocationStatus.FAILURE`. Defaults to None.
    """

    status: ShutdownRepositoryLocationStatus
    message: str | None = None


class JobInfo(NamedTuple):
    repository_location_name: str
    repository_name: str
    job_name: str

    @staticmethod
    def from_node(node: dict[str, Any]) -> list["JobInfo"]:
        repo_name = node["name"]
        repo_location_name = node["location"]["name"]
        return [
            JobInfo(
                repository_location_name=repo_location_name,
                repository_name=repo_name,
                job_name=job["name"],
            )
            for job in node["pipelines"]
        ]


class InvalidOutputErrorInfo(NamedTuple):
    """This class gives information about an InvalidOutputError from submitting a pipeline for execution
    from GraphQL.

    Args:
        step_key (str): key of the step that failed
        invalid_output_name (str): the name of the invalid output from the given step
    """

    step_key: str
    invalid_output_name: str



import asyncio

import graphene
from dagster._core.definitions.asset_health.asset_check_health import (
    AssetHealthCheckDegradedMetadata,
    AssetHealthCheckMetadata,
    AssetHealthCheckUnknownMetadata,
    AssetHealthCheckWarningMetadata,
    get_asset_check_status_and_metadata,
)
from dagster._core.definitions.asset_health.asset_freshness_health import (
    get_freshness_status_and_metadata,
)
from dagster._core.definitions.asset_health.asset_health import (
    AssetHealthStatus,
    overall_status_from_component_statuses,
)
from dagster._core.definitions.asset_health.asset_materialization_health import (
    AssetHealthMaterializationDegradedNotPartitionedMeta,
    AssetHealthMaterializationDegradedPartitionedMeta,
    AssetHealthMaterializationHealthyPartitionedMeta,
    AssetHealthMaterializationMetadata,
    get_materialization_status_and_metadata,
)

from dagster_graphql.schema.entity_key import GrapheneAssetKey
from dagster_graphql.schema.util import ResolveInfo

GrapheneAssetHealthStatus = graphene.Enum.from_enum(AssetHealthStatus)


class GrapheneAssetHealthCheckDegradedMeta(graphene.ObjectType):
    numFailedChecks = graphene.NonNull(graphene.Int)
    numWarningChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckDegradedMeta"


class GrapheneAssetHealthCheckWarningMeta(graphene.ObjectType):
    numWarningChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckWarningMeta"


class GrapheneAssetHealthCheckUnknownMeta(graphene.ObjectType):
    numNotExecutedChecks = graphene.NonNull(graphene.Int)
    totalNumChecks = graphene.NonNull(graphene.Int)

    class Meta:
        name = "AssetHealthCheckUnknownMeta"


class GrapheneAssetHealthCheckMeta(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetHealthCheckDegradedMeta,
            GrapheneAssetHealthCheckWarningMeta,
            GrapheneAssetHealthCheckUnknownMeta,
        )
        name = "AssetHealthCheckMeta"

    @staticmethod
    def from_metadata_class(
        metadata: AssetHealthCheckMetadata,
    ) -> "GrapheneAssetHealthCheckMeta":
        if isinstance(metadata, AssetHealthCheckDegradedMetadata):
            return GrapheneAssetHealthCheckDegradedMeta(
                numFailedChecks=metadata.num_failed_checks,
                numWarningChecks=metadata.num_warning_checks,
                totalNumChecks=metadata.total_num_checks,
            )
        elif isinstance(metadata, AssetHealthCheckWarningMetadata):
            return GrapheneAssetHealthCheckWarningMeta(
                numWarningChecks=metadata.num_warning_checks,
                totalNumChecks=metadata.total_num_checks,
            )
        elif isinstance(metadata, AssetHealthCheckUnknownMetadata):
            return GrapheneAssetHealthCheckUnknownMeta(
                numNotExecutedChecks=metadata.num_not_executed_checks,
                totalNumChecks=metadata.total_num_checks,
            )
        else:
            raise ValueError(f"Unknown metadata class: {type(metadata)}")


class GrapheneAssetHealthMaterializationDegradedPartitionedMeta(graphene.ObjectType):
    numFailedPartitions = graphene.NonNull(graphene.Int)
    numMissingPartitions = graphene.NonNull(graphene.Int)
    totalNumPartitions = graphene.NonNull(graphene.Int)
    latestRunId = graphene.String()
    latestFailedRunId = graphene.String()

    class Meta:
        name = "AssetHealthMaterializationDegradedPartitionedMeta"


class GrapheneAssetHealthMaterializationHealthyPartitionedMeta(graphene.ObjectType):
    numMissingPartitions = graphene.NonNull(graphene.Int)
    totalNumPartitions = graphene.NonNull(graphene.Int)
    latestRunId = graphene.String()

    class Meta:
        name = "AssetHealthMaterializationHealthyPartitionedMeta"


class GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta(graphene.ObjectType):
    failedRunId = graphene.String()

    class Meta:
        name = "AssetHealthMaterializationDegradedNotPartitionedMeta"


class GrapheneAssetHealthMaterializationMeta(graphene.Union):
    class Meta:
        types = (
            GrapheneAssetHealthMaterializationDegradedPartitionedMeta,
            GrapheneAssetHealthMaterializationHealthyPartitionedMeta,
            GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta,
        )
        name = "AssetHealthMaterializationMeta"

    @staticmethod
    def from_metadata_class(
        asset_key: GrapheneAssetKey,
        metadata: AssetHealthMaterializationMetadata,
    ) -> "GrapheneAssetHealthMaterializationMeta":
        if isinstance(metadata, AssetHealthMaterializationDegradedNotPartitionedMeta):
            return GrapheneAssetHealthMaterializationDegradedNotPartitionedMeta(
                failedRunId=metadata.failed_run_id,
            )
        elif isinstance(metadata, AssetHealthMaterializationHealthyPartitionedMeta):
            return GrapheneAssetHealthMaterializationHealthyPartitionedMeta(
                numMissingPartitions=metadata.num_missing_partitions,
                totalNumPartitions=metadata.total_num_partitions,
                latestRunId=metadata.latest_run_id,
            )
        elif isinstance(metadata, AssetHealthMaterializationDegradedPartitionedMeta):
            return GrapheneAssetHealthMaterializationDegradedPartitionedMeta(
                numFailedPartitions=metadata.num_failed_partitions,
                numMissingPartitions=metadata.num_missing_partitions,
                totalNumPartitions=metadata.total_num_partitions,
                latestRunId=metadata.latest_run_id,
                latestFailedRunId=metadata.latest_failed_to_materialize_run_id,
            )
        else:
            raise ValueError(f"Unknown metadata class: {type(metadata)}")


class GrapheneAssetHealthFreshnessMeta(graphene.ObjectType):
    lastMaterializedTimestamp = graphene.Field(graphene.Float)

    class Meta:
        name = "AssetHealthFreshnessMeta"


class GrapheneAssetHealth(graphene.ObjectType):
    assetHealth = graphene.NonNull(GrapheneAssetHealthStatus)
    materializationStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    materializationStatusMetadata = graphene.Field(GrapheneAssetHealthMaterializationMeta)
    assetChecksStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    assetChecksStatusMetadata = graphene.Field(GrapheneAssetHealthCheckMeta)
    freshnessStatus = graphene.NonNull(GrapheneAssetHealthStatus)
    freshnessStatusMetadata = graphene.Field(GrapheneAssetHealthFreshnessMeta)

    class Meta:
        name = "AssetHealth"

    def __init__(self, asset_key: GrapheneAssetKey, dynamic_partitions_loader):
        super().__init__()
        self._asset_key = asset_key
        self._dynamic_partitions_loader = dynamic_partitions_loader
        self.materialization_status_task = None
        self.asset_check_status_task = None
        self.freshness_status_task = None

    async def resolve_materializationStatus(self, graphene_info: ResolveInfo) -> AssetHealthStatus:
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                get_materialization_status_and_metadata(graphene_info.context, self._asset_key)
            )
        materialization_status, _ = await self.materialization_status_task
        return materialization_status

    async def resolve_materializationStatusMetadata(
        self, graphene_info: ResolveInfo
    ) -> GrapheneAssetHealthMaterializationMeta:
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                get_materialization_status_and_metadata(graphene_info.context, self._asset_key)
            )
        _, materialization_status_metadata = await self.materialization_status_task
        return (
            GrapheneAssetHealthMaterializationMeta.from_metadata_class(
                asset_key=self._asset_key, metadata=materialization_status_metadata
            )
            if materialization_status_metadata
            else None
        )

    async def resolve_assetChecksStatus(self, graphene_info: ResolveInfo) -> AssetHealthStatus:
        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                get_asset_check_status_and_metadata(graphene_info.context, self._asset_key)
            )

        asset_checks_status, _ = await self.asset_check_status_task
        return asset_checks_status

    async def resolve_assetChecksStatusMetadata(
        self, graphene_info: ResolveInfo
    ) -> GrapheneAssetHealthCheckMeta:
        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                get_asset_check_status_and_metadata(graphene_info.context, self._asset_key)
            )

        _, asset_checks_status_metadata = await self.asset_check_status_task
        return (
            GrapheneAssetHealthCheckMeta.from_metadata_class(asset_checks_status_metadata)
            if asset_checks_status_metadata
            else None
        )

    async def resolve_freshnessStatus(self, graphene_info: ResolveInfo) -> AssetHealthStatus:
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                get_freshness_status_and_metadata(graphene_info.context, self._asset_key)
            )

        freshness_status, _ = await self.freshness_status_task
        return freshness_status

    async def resolve_freshnessStatusMetadata(
        self, graphene_info: ResolveInfo
    ) -> GrapheneAssetHealthFreshnessMeta:
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                get_freshness_status_and_metadata(graphene_info.context, self._asset_key)
            )

        _, freshness_status_metadata = await self.freshness_status_task
        return (
            GrapheneAssetHealthFreshnessMeta(
                lastMaterializedTimestamp=freshness_status_metadata.last_materialized_timestamp
            )
            if freshness_status_metadata
            else None
        )

    async def resolve_assetHealth(self, graphene_info: ResolveInfo) -> AssetHealthStatus:
        if not graphene_info.context.instance.dagster_asset_health_queries_supported():
            return AssetHealthStatus.UNKNOWN

        if self.asset_check_status_task is None:
            self.asset_check_status_task = asyncio.create_task(
                get_asset_check_status_and_metadata(graphene_info.context, self._asset_key)
            )
        if self.materialization_status_task is None:
            self.materialization_status_task = asyncio.create_task(
                get_materialization_status_and_metadata(graphene_info.context, self._asset_key)
            )
        if self.freshness_status_task is None:
            self.freshness_status_task = asyncio.create_task(
                get_freshness_status_and_metadata(graphene_info.context, self._asset_key)
            )

        asset_checks_status, _ = await self.asset_check_status_task
        materialization_status, _ = await self.materialization_status_task
        freshness_status, _ = await self.freshness_status_task

        return overall_status_from_component_statuses(
            asset_checks_status=asset_checks_status,
            materialization_status=materialization_status,
            freshness_status=freshness_status,
        )



from collections.abc import Callable

import dagster._check as check
import graphene
from dagster._core.snap import ConfigTypeSnap, LoggerDefSnap

from dagster_graphql.schema.config_types import GrapheneConfigTypeField
from dagster_graphql.schema.util import ResolveInfo


class GrapheneLogger(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    configField = graphene.Field(GrapheneConfigTypeField)

    class Meta:
        name = "Logger"

    def __init__(
        self,
        get_config_type: Callable[[str], ConfigTypeSnap],
        logger_def_snap: LoggerDefSnap,
    ):
        super().__init__()
        self._get_config_type = get_config_type
        self._logger_def_snap = check.inst_param(logger_def_snap, "logger_def_snap", LoggerDefSnap)
        self.name = logger_def_snap.name
        self.description = logger_def_snap.description

    def resolve_configField(self, _: ResolveInfo):
        if self._logger_def_snap.config_field_snap:
            try:
                # config type may not be present if mode config mapped, null out gracefully
                self._get_config_type(self._logger_def_snap.config_field_snap.type_key)
            except KeyError:
                return None

            return GrapheneConfigTypeField(
                self._get_config_type,
                field_snap=self._logger_def_snap.config_field_snap,
            )

        return None



import graphene

from dagster_graphql.schema.config_types import (
    GrapheneCompositeConfigType,
    GrapheneEnumConfigType,
    GrapheneRegularConfigType,
)
from dagster_graphql.schema.errors import (
    GrapheneConfigTypeNotFoundError,
    GraphenePipelineNotFoundError,
    GraphenePythonError,
)


class GrapheneConfigTypeOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneEnumConfigType,
            GrapheneCompositeConfigType,
            GrapheneRegularConfigType,
            GraphenePipelineNotFoundError,
            GrapheneConfigTypeNotFoundError,
            GraphenePythonError,
        )
        name = "ConfigTypeOrError"



import graphene
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.events import DagsterEventType
from dagster._core.execution.backfill import BulkActionsFilter, BulkActionStatus
from dagster._core.storage.dagster_run import DagsterRunStatus, RunsFilter
from dagster._time import datetime_from_timestamp
from dagster._utils import check

from dagster_graphql.schema.asset_checks import GrapheneAssetCheckSeverity
from dagster_graphql.schema.pipelines.status import GrapheneRunStatus
from dagster_graphql.schema.runs import GrapheneRunConfigData
from dagster_graphql.schema.util import non_null_list


class GrapheneAssetKeyInput(graphene.InputObjectType):
    path = non_null_list(graphene.String)

    class Meta:
        name = "AssetKeyInput"

    def to_asset_key(self) -> AssetKey:
        return AssetKey(self.path)


class GrapheneAssetCheckHandleInput(graphene.InputObjectType):
    assetKey = graphene.NonNull(GrapheneAssetKeyInput)
    name = graphene.NonNull(graphene.String)

    class Meta:
        name = "AssetCheckHandleInput"


class GrapheneExecutionTag(graphene.InputObjectType):
    key = graphene.NonNull(graphene.String)
    value = graphene.NonNull(graphene.String)

    class Meta:
        name = "ExecutionTag"


class GrapheneRunsFilter(graphene.InputObjectType):
    runIds = graphene.List(graphene.String)
    pipelineName = graphene.InputField(graphene.String)
    tags = graphene.List(graphene.NonNull(GrapheneExecutionTag))
    statuses = graphene.List(graphene.NonNull(GrapheneRunStatus))
    snapshotId = graphene.InputField(graphene.String)
    updatedAfter = graphene.InputField(graphene.Float)
    updatedBefore = graphene.InputField(graphene.Float)
    createdBefore = graphene.InputField(graphene.Float)
    createdAfter = graphene.InputField(graphene.Float)
    mode = graphene.InputField(graphene.String)

    class Meta:
        description = """This type represents a filter on Dagster runs."""
        name = "RunsFilter"

    def to_selector(self):
        if self.tags:
            # We are wrapping self.tags in a list because graphene.List is not marked as iterable
            tags = {tag["key"]: tag["value"] for tag in list(self.tags)}
        else:
            tags = None

        if self.statuses:
            statuses = [DagsterRunStatus[status.value] for status in self.statuses]
        else:
            statuses = None

        updated_before = datetime_from_timestamp(self.updatedBefore) if self.updatedBefore else None
        updated_after = datetime_from_timestamp(self.updatedAfter) if self.updatedAfter else None
        created_before = datetime_from_timestamp(self.createdBefore) if self.createdBefore else None
        created_after = datetime_from_timestamp(self.createdAfter) if self.createdAfter else None

        return RunsFilter(
            run_ids=self.runIds if self.runIds else None,
            job_name=self.pipelineName,
            tags=tags,
            statuses=statuses,
            snapshot_id=self.snapshotId,
            updated_before=updated_before,
            updated_after=updated_after,
            created_before=created_before,
            created_after=created_after,
        )


class GrapheneStepOutputHandle(graphene.InputObjectType):
    stepKey = graphene.NonNull(graphene.String)
    outputName = graphene.NonNull(graphene.String)

    class Meta:
        name = "StepOutputHandle"


class GraphenePipelineSelector(graphene.InputObjectType):
    pipelineName = graphene.NonNull(graphene.String)
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))
    assetSelection = graphene.List(graphene.NonNull(GrapheneAssetKeyInput))
    assetCheckSelection = graphene.List(graphene.NonNull(GrapheneAssetCheckHandleInput))

    class Meta:
        description = """This type represents the fields necessary to identify a
        pipeline or pipeline subset."""
        name = "PipelineSelector"


class GrapheneAssetGroupSelector(graphene.InputObjectType):
    groupName = graphene.NonNull(graphene.String)
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        description = """This type represents the fields necessary to identify
        an asset group."""
        name = "AssetGroupSelector"


class GrapheneGraphSelector(graphene.InputObjectType):
    graphName = graphene.NonNull(graphene.String)
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        description = """This type represents the fields necessary to identify a
        graph"""
        name = "GraphSelector"


class GrapheneJobOrPipelineSelector(graphene.InputObjectType):
    pipelineName = graphene.String()
    jobName = graphene.String()
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)
    solidSelection = graphene.List(graphene.NonNull(graphene.String))
    assetSelection = graphene.List(graphene.NonNull(GrapheneAssetKeyInput))
    assetCheckSelection = graphene.List(graphene.NonNull(GrapheneAssetCheckHandleInput))

    class Meta:
        description = """This type represents the fields necessary to identify a job or pipeline"""
        name = "JobOrPipelineSelector"


class GrapheneRepositorySelector(graphene.InputObjectType):
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)

    class Meta:
        description = """This type represents the fields necessary to identify a repository."""
        name = "RepositorySelector"


class GraphenePartitionSetSelector(graphene.InputObjectType):
    partitionSetName = graphene.NonNull(graphene.String)
    repositorySelector = graphene.NonNull(GrapheneRepositorySelector)

    class Meta:
        description = """This type represents the fields necessary to identify a
        pipeline or pipeline subset."""
        name = "PartitionSetSelector"


class GraphenePartitionRangeSelector(graphene.InputObjectType):
    start = graphene.NonNull(graphene.String)
    end = graphene.NonNull(graphene.String)

    class Meta:
        description = """This type represents a partition range selection with start and end."""
        name = "PartitionRangeSelector"


class GraphenePartitionsSelector(graphene.InputObjectType):
    range = graphene.InputField(GraphenePartitionRangeSelector)
    ranges = graphene.InputField(graphene.List(graphene.NonNull(GraphenePartitionRangeSelector)))

    class Meta:
        description = """This type represents a partitions selection."""
        name = "PartitionsSelector"


class GraphenePartitionsByAssetSelector(graphene.InputObjectType):
    assetKey = graphene.NonNull(GrapheneAssetKeyInput)
    partitions = graphene.InputField(GraphenePartitionsSelector)

    class Meta:
        description = """This type represents a partitions selection for an asset."""
        name = "PartitionsByAssetSelector"


class GrapheneAssetBackfillPreviewParams(graphene.InputObjectType):
    partitionNames = graphene.InputField(non_null_list(graphene.String))
    assetSelection = graphene.InputField(non_null_list(GrapheneAssetKeyInput))

    class Meta:
        name = "AssetBackfillPreviewParams"


class GrapheneLaunchBackfillParams(graphene.InputObjectType):
    selector = graphene.InputField(GraphenePartitionSetSelector)
    partitionNames = graphene.List(graphene.NonNull(graphene.String))
    partitionsByAssets = graphene.List(GraphenePartitionsByAssetSelector)
    reexecutionSteps = graphene.List(graphene.NonNull(graphene.String))
    assetSelection = graphene.InputField(graphene.List(graphene.NonNull(GrapheneAssetKeyInput)))
    fromFailure = graphene.Boolean()
    allPartitions = graphene.Boolean()
    tags = graphene.List(graphene.NonNull(GrapheneExecutionTag))
    forceSynchronousSubmission = graphene.Boolean()
    title = graphene.String()
    description = graphene.String()
    runConfigData = graphene.InputField(GrapheneRunConfigData)

    class Meta:
        name = "LaunchBackfillParams"


class GrapheneRunlessAssetEventType(graphene.Enum):
    """The event type of an asset event."""

    ASSET_MATERIALIZATION = "ASSET_MATERIALIZATION"
    ASSET_OBSERVATION = "ASSET_OBSERVATION"

    class Meta:
        name = "AssetEventType"

    def to_dagster_event_type(self) -> DagsterEventType:
        if self == GrapheneRunlessAssetEventType.ASSET_MATERIALIZATION:
            return DagsterEventType.ASSET_MATERIALIZATION
        elif self == GrapheneRunlessAssetEventType.ASSET_OBSERVATION:
            return DagsterEventType.ASSET_OBSERVATION
        else:
            check.failed(f"unhandled type {self}")


class GrapheneReportRunlessAssetEventsParams(graphene.InputObjectType):
    eventType = graphene.NonNull(GrapheneRunlessAssetEventType)
    assetKey = graphene.NonNull(GrapheneAssetKeyInput)
    partitionKeys = graphene.InputField(graphene.List(graphene.String))
    description = graphene.String()

    class Meta:
        name = "ReportRunlessAssetEventsParams"


class GrapheneReportAssetCheckEvaluationParams(graphene.InputObjectType):
    assetKey = graphene.NonNull(GrapheneAssetKeyInput)
    checkName = graphene.NonNull(graphene.String)
    passed = graphene.NonNull(graphene.Boolean)
    severity = graphene.InputField(GrapheneAssetCheckSeverity)
    serializedMetadata = graphene.InputField(graphene.String)
    partition = graphene.InputField(graphene.String)

    class Meta:
        name = "ReportAssetCheckEvaluationParams"


class GrapheneSensorSelector(graphene.InputObjectType):
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)
    sensorName = graphene.NonNull(graphene.String)

    class Meta:
        description = """This type represents the fields necessary to identify a sensor."""
        name = "SensorSelector"


class GrapheneScheduleSelector(graphene.InputObjectType):
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)
    scheduleName = graphene.NonNull(graphene.String)

    class Meta:
        description = """This type represents the fields necessary to identify a schedule."""
        name = "ScheduleSelector"


class GrapheneResourceSelector(graphene.InputObjectType):
    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)
    resourceName = graphene.NonNull(graphene.String)

    class Meta:
        description = (
            """This type represents the fields necessary to identify a top-level resource."""
        )
        name = "ResourceSelector"


class GrapheneExecutionMetadata(graphene.InputObjectType):
    tags = graphene.List(graphene.NonNull(GrapheneExecutionTag))
    rootRunId = graphene.String(
        description="""The ID of the run at the root of the run group. All partial /
        full re-executions should use the first run as the rootRunID so they are
        grouped together."""
    )
    parentRunId = graphene.String(
        description="""The ID of the run serving as the parent within the run group.
        For the first re-execution, this will be the same as the `rootRunId`. For
        subsequent runs, the root or a previous re-execution could be the parent run."""
    )

    class Meta:
        name = "ExecutionMetadata"


class GrapheneExecutionParams(graphene.InputObjectType):
    selector = graphene.NonNull(
        GrapheneJobOrPipelineSelector,
        description="""Defines the job / pipeline and solid subset that should be executed.
        All subsequent executions in the same run group (for example, a single-step
        re-execution) are scoped to the original run's selector and solid
        subset.""",
    )
    runConfigData = graphene.InputField(GrapheneRunConfigData)
    mode = graphene.InputField(graphene.String)
    executionMetadata = graphene.InputField(
        GrapheneExecutionMetadata,
        description="""Defines run tags and parent / root relationships.\n\nNote: To
        'restart from failure', provide a `parentRunId` and pass the
        'dagster/is_resume_retry' tag. Dagster's automatic step key selection will
        override any stepKeys provided.""",
    )
    stepKeys = graphene.InputField(
        graphene.List(graphene.NonNull(graphene.String)),
        description="""Defines step keys to execute within the execution plan defined
        by the pipeline `selector`. To execute the entire execution plan, you can omit
        this parameter, provide an empty array, or provide every step name.""",
    )
    preset = graphene.InputField(graphene.String)

    class Meta:
        name = "ExecutionParams"


class GrapheneReexecutionStrategy(graphene.Enum):
    FROM_ASSET_FAILURE = "FROM_ASSET_FAILURE"
    FROM_FAILURE = "FROM_FAILURE"
    ALL_STEPS = "ALL_STEPS"

    class Meta:
        name = "ReexecutionStrategy"


class GrapheneReexecutionParams(graphene.InputObjectType):
    parentRunId = graphene.NonNull(graphene.String)
    strategy = graphene.NonNull(GrapheneReexecutionStrategy)
    extraTags = graphene.List(
        graphene.NonNull(GrapheneExecutionTag),
        description="""When re-executing a single run, pass new tags which will upsert over tags on the parent run.""",
    )
    useParentRunTags = graphene.Boolean(
        description="""When re-executing a single run, pass false to avoid adding the parent run tags by default."""
    )

    class Meta:
        name = "ReexecutionParams"


class GrapheneMarshalledInput(graphene.InputObjectType):
    input_name = graphene.NonNull(graphene.String)
    key = graphene.NonNull(graphene.String)

    class Meta:
        name = "MarshalledInput"


class GrapheneMarshalledOutput(graphene.InputObjectType):
    output_name = graphene.NonNull(graphene.String)
    key = graphene.NonNull(graphene.String)

    class Meta:
        name = "MarshalledOutput"


class GrapheneStepExecution(graphene.InputObjectType):
    stepKey = graphene.NonNull(graphene.String)
    marshalledInputs = graphene.List(graphene.NonNull(GrapheneMarshalledInput))
    marshalledOutputs = graphene.List(graphene.NonNull(GrapheneMarshalledOutput))

    class Meta:
        name = "StepExecution"


class GrapheneInstigationSelector(graphene.InputObjectType):
    class Meta:
        name = "InstigationSelector"
        description = (
            """This type represents the fields necessary to identify a schedule or sensor."""
        )

    repositoryName = graphene.NonNull(graphene.String)
    repositoryLocationName = graphene.NonNull(graphene.String)
    name = graphene.NonNull(graphene.String)


class GrapheneTagInput(graphene.InputObjectType):
    key = graphene.NonNull(graphene.String)
    value = graphene.NonNull(graphene.String)

    class Meta:
        name = "TagInput"


class GrapheneBulkActionsFilter(graphene.InputObjectType):
    statuses = graphene.List(
        graphene.NonNull("dagster_graphql.schema.backfill.GrapheneBulkActionStatus")
    )
    createdBefore = graphene.InputField(graphene.Float)
    createdAfter = graphene.InputField(graphene.Float)

    class Meta:
        description = """This type represents a filter on Dagster Bulk Actions (backfills)."""
        name = "BulkActionsFilter"

    def to_selector(self):
        statuses = (
            [BulkActionStatus[status.value] for status in self.statuses] if self.statuses else None
        )
        created_before = datetime_from_timestamp(self.createdBefore) if self.createdBefore else None
        created_after = datetime_from_timestamp(self.createdAfter) if self.createdAfter else None

        return BulkActionsFilter(
            statuses=statuses,
            created_before=created_before,
            created_after=created_after,
        )


types = [
    GrapheneAssetKeyInput,
    GrapheneExecutionMetadata,
    GrapheneExecutionParams,
    GrapheneExecutionTag,
    GrapheneInstigationSelector,
    GrapheneMarshalledInput,
    GrapheneMarshalledOutput,
    GrapheneLaunchBackfillParams,
    GraphenePartitionSetSelector,
    GraphenePartitionsByAssetSelector,
    GrapheneRunsFilter,
    GraphenePipelineSelector,
    GrapheneRepositorySelector,
    GrapheneResourceSelector,
    GrapheneScheduleSelector,
    GrapheneSensorSelector,
    GrapheneStepExecution,
    GrapheneStepOutputHandle,
    GrapheneTagInput,
    GrapheneReportRunlessAssetEventsParams,
    GrapheneBulkActionsFilter,
]



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import graphene
from dagster._core.definitions.selector import SensorSelector

from dagster_graphql.implementation.utils import capture_error
from dagster_graphql.schema.errors import GraphenePythonError, GrapheneSensorNotFoundError
from dagster_graphql.schema.inputs import GrapheneSensorSelector
from dagster_graphql.schema.instigation import GrapheneDryRunInstigationTick

if TYPE_CHECKING:
    from dagster_graphql.schema.util import ResolveInfo


class GrapheneSensorDryRunResult(graphene.Union):
    class Meta:
        types = (GraphenePythonError, GrapheneSensorNotFoundError, GrapheneDryRunInstigationTick)
        name = "SensorDryRunResult"


class GrapheneSensorDryRunMutation(graphene.Mutation):
    """Enable a sensor to launch runs for a job based on external state change."""

    Output = graphene.NonNull(GrapheneSensorDryRunResult)

    class Arguments:
        selector_data = graphene.NonNull(GrapheneSensorSelector)
        cursor = graphene.String()

    class Meta:
        name = "SensorDryRunMutation"

    @capture_error
    def mutate(
        self, graphene_info: "ResolveInfo", selector_data: Mapping[str, Any], cursor: str | None
    ):
        return GrapheneDryRunInstigationTick(
            SensorSelector.from_graphql_input(selector_data), timestamp=None, cursor=cursor
        )



import graphene

from dagster_graphql.schema.pipelines.pipeline import GraphenePipeline
from dagster_graphql.schema.solids import GrapheneISolidDefinition, GrapheneSolidHandle
from dagster_graphql.schema.util import non_null_list


class GrapheneNodeInvocationSite(graphene.ObjectType):
    class Meta:
        description = """An invocation of a solid within a repo."""
        name = "NodeInvocationSite"

    pipeline = graphene.NonNull(GraphenePipeline)
    solidHandle = graphene.NonNull(GrapheneSolidHandle)


class GrapheneUsedSolid(graphene.ObjectType):
    class Meta:
        description = """A solid definition and its invocations within the repo."""
        name = "UsedSolid"

    definition = graphene.NonNull(GrapheneISolidDefinition)
    invocations = non_null_list(GrapheneNodeInvocationSite)



import random
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any
from unittest.mock import PropertyMock, patch

import dagster._check as check
from dagster import AssetKey, AutomationCondition, RunRequest, asset, evaluate_automation_conditions
from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.asset_daemon_cursor import AssetDaemonCursor
from dagster._core.definitions.declarative_automation.operators.since_operator import (
    SinceConditionData,
)
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionEvaluation,
    AutomationConditionEvaluationWithRunIds,
    AutomationConditionNodeSnapshot,
    HistoricalAllPartitionsSubsetSentinel,
)
from dagster._core.definitions.partitions.definition import (
    PartitionsDefinition,
    StaticPartitionsDefinition,
)
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.definitions.sensor_definition import SensorType
from dagster._core.instance import DagsterInstance
from dagster._core.remote_origin import RemoteInstigatorOrigin
from dagster._core.scheduler.instigation import (
    InstigatorState,
    InstigatorStatus,
    SensorInstigatorData,
    TickData,
    TickStatus,
)
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster._daemon.asset_daemon import (
    _PRE_SENSOR_AUTO_MATERIALIZE_INSTIGATOR_NAME,
    _PRE_SENSOR_AUTO_MATERIALIZE_ORIGIN_ID,
    _PRE_SENSOR_AUTO_MATERIALIZE_SELECTOR_ID,
    asset_daemon_cursor_to_instigator_serialized_cursor,
)
from dagster._time import get_current_datetime
from dagster._vendored.dateutil.relativedelta import relativedelta
from dagster_graphql.test.utils import execute_dagster_graphql, infer_repository

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    ExecutingGraphQLContextTestMatrix,
)

if TYPE_CHECKING:
    from dagster._core.definitions.asset_key import EntityKey

TICKS_QUERY = """
query AssetDameonTicksQuery($dayRange: Int, $dayOffset: Int, $statuses: [InstigationTickStatus!], $limit: Int, $cursor: String, $beforeTimestamp: Float, $afterTimestamp: Float) {
    autoMaterializeTicks(dayRange: $dayRange, dayOffset: $dayOffset, statuses: $statuses, limit: $limit, cursor: $cursor, beforeTimestamp: $beforeTimestamp, afterTimestamp: $afterTimestamp) {
        id
        timestamp
        endTimestamp
        status
        requestedAssetKeys {
            path
        }
        requestedMaterializationsForAssets {
            assetKey {
                path
            }
            partitionKeys
        }
        requestedAssetMaterializationCount
        autoMaterializeAssetEvaluationId
    }
}
"""


def _create_tick(instance, status, timestamp, evaluation_id, run_requests=None, end_timestamp=None):
    return instance.create_tick(
        TickData(
            instigator_origin_id=_PRE_SENSOR_AUTO_MATERIALIZE_ORIGIN_ID,
            instigator_name=_PRE_SENSOR_AUTO_MATERIALIZE_INSTIGATOR_NAME,
            instigator_type=InstigatorType.AUTO_MATERIALIZE,
            status=status,
            timestamp=timestamp,
            end_timestamp=end_timestamp,
            selector_id=_PRE_SENSOR_AUTO_MATERIALIZE_SELECTOR_ID,
            run_ids=[],
            auto_materialize_evaluation_id=evaluation_id,
            run_requests=run_requests,
        )
    )


class TestAutoMaterializeTicks(ExecutingGraphQLContextTestMatrix):
    def test_get_tick_range(self, graphql_context):
        result = execute_dagster_graphql(
            graphql_context,
            TICKS_QUERY,
            variables={"dayRange": None, "dayOffset": None},
        )
        assert len(result.data["autoMaterializeTicks"]) == 0

        now = get_current_datetime()
        end_timestamp = now.timestamp() + 20

        success_1 = _create_tick(
            graphql_context.instance,
            TickStatus.SUCCESS,
            now.timestamp(),
            end_timestamp=end_timestamp,
            evaluation_id=3,
            run_requests=[
                RunRequest(asset_selection=[AssetKey("foo"), AssetKey("bar")], partition_key="abc"),
                RunRequest(asset_selection=[AssetKey("bar")], partition_key="def"),
                RunRequest(asset_selection=[AssetKey("baz")], partition_key=None),
            ],
        )

        success_2 = _create_tick(
            graphql_context.instance,
            TickStatus.SUCCESS,
            (now - relativedelta(days=1, hours=1)).timestamp(),
            evaluation_id=2,
        )

        _create_tick(
            graphql_context.instance,
            TickStatus.SKIPPED,
            (now - relativedelta(days=2, hours=1)).timestamp(),
            evaluation_id=1,
        )

        result = execute_dagster_graphql(
            graphql_context,
            TICKS_QUERY,
            variables={"dayRange": None, "dayOffset": None},
        )
        assert len(result.data["autoMaterializeTicks"]) == 3

        result = execute_dagster_graphql(
            graphql_context,
            TICKS_QUERY,
            variables={"dayRange": 1, "dayOffset": None},
        )
        assert len(result.data["autoMaterializeTicks"]) == 1
        tick = result.data["autoMaterializeTicks"][0]
        assert tick["endTimestamp"] == end_timestamp
        assert tick["autoMaterializeAssetEvaluationId"] == "3"
        assert sorted(tick["requestedAssetKeys"], key=lambda x: x["path"][0]) == [
            {"path": ["bar"]},
            {"path": ["baz"]},
            {"path": ["foo"]},
        ]

        asset_materializations = tick["requestedMaterializationsForAssets"]
        by_asset_key = {
            AssetKey.from_coercible(mat["assetKey"]["path"]).to_user_string(): mat["partitionKeys"]
            for mat in asset_materializations
        }

        assert {key: sorted(val) for key, val in by_asset_key.items()} == {
            "foo": ["abc"],
            "bar": ["abc", "def"],
            "baz": [],
        }

        assert tick["requestedAssetMaterializationCount"] == 4

        result = execute_dagster_graphql(
            graphql_context,
            TICKS_QUERY,
            variables={
                "beforeTimestamp": success_2.timestamp + 1,
                "afterTimestamp": success_2.timestamp - 1,
            },
        )
        assert len(result.data["autoMaterializeTicks"]) == 1
        tick = result.data["autoMaterializeTicks"][0]
        assert tick["autoMaterializeAssetEvaluationId"] == str(
            success_2.automation_condition_evaluation_id
        )

        result = execute_dagster_graphql(
            graphql_context,
            TICKS_QUERY,
            variables={"dayRange": None, "dayOffset": None, "statuses": ["SUCCESS"]},
        )
        assert len(result.data["autoMaterializeTicks"]) == 2

        result = execute_dagster_graphql(
            graphql_context,
            TICKS_QUERY,
            variables={"dayRange": None, "dayOffset": None, "statuses": ["SUCCESS"], "limit": 1},
        )
        ticks = result.data["autoMaterializeTicks"]
        assert len(ticks) == 1
        assert ticks[0]["timestamp"] == success_1.timestamp
        assert ticks[0]["autoMaterializeAssetEvaluationId"] == str(
            success_1.automation_condition_evaluation_id
        )

        cursor = ticks[0]["id"]

        result = execute_dagster_graphql(
            graphql_context,
            TICKS_QUERY,
            variables={
                "dayRange": None,
                "dayOffset": None,
                "statuses": ["SUCCESS"],
                "limit": 1,
                "cursor": cursor,
            },
        )
        ticks = result.data["autoMaterializeTicks"]
        assert len(ticks) == 1
        assert ticks[0]["timestamp"] == success_2.timestamp


ENTITY_FRAGMENT = """
fragment entityKeyFragment on EntityKey {
    ... on AssetKey {
        path
    }
    ... on AssetCheckhandle {
        name
        assetKey {
            path
        }
    }
}
"""
FRAGMENTS = (
    ENTITY_FRAGMENT
    + """
fragment evaluationFields on AssetConditionEvaluation {
    rootUniqueId
    evaluationNodes {
        ... on UnpartitionedAssetConditionEvaluationNode {
            description
            startTimestamp
            endTimestamp
            status
            uniqueId
            childUniqueIds
            entityKey {
                ...entityKeyFragment
            }
        }
        ... on PartitionedAssetConditionEvaluationNode {
            description
            startTimestamp
            endTimestamp
            numTrue
            uniqueId
            childUniqueIds
            entityKey {
                ...entityKeyFragment
            }
        }
        ... on SpecificPartitionAssetConditionEvaluationNode {
            description
            status
            uniqueId
            childUniqueIds
            entityKey {
                ...entityKeyFragment
            }
        }
    }
}
"""
)

AUTO_MATERIALIZE_POLICY_SENSORS_QUERY = """
query GetEvaluationsQuery($assetKey: AssetKeyInput!) {
    assetNodeOrError(assetKey: $assetKey) {
        ... on AssetNode {
            currentAutoMaterializeEvaluationId
            targetingInstigators {
                ... on Schedule {
                    name
                }
                ... on Sensor {
                    name
                }
            }
        }
    }
}
"""


LEGACY_QUERY = (
    FRAGMENTS
    + """
query GetEvaluationsQuery($assetKey: AssetKeyInput!, $limit: Int!, $cursor: String) {
    assetNodeOrError(assetKey: $assetKey) {
        ... on AssetNode {
            currentAutoMaterializeEvaluationId
        }
    }
    assetConditionEvaluationRecordsOrError(assetKey: $assetKey, limit: $limit, cursor: $cursor) {
        ... on AssetConditionEvaluationRecords {
            records {
                evaluationId
                numRequested
                assetKey {
                    path
                }
                evaluation {
                    ...evaluationFields
                }
            }
        }
    }
}
"""
)

LEGACY_QUERY_FOR_SPECIFIC_PARTITION = (
    FRAGMENTS
    + """
query GetPartitionEvaluationQuery($assetKey: AssetKeyInput!, $partition: String!, $evaluationId: ID!) {
    assetConditionEvaluationForPartition(assetKey: $assetKey, partition: $partition, evaluationId: $evaluationId) {
        ...evaluationFields
    }
}
"""
)

LEGACY_QUERY_FOR_EVALUATION_ID = (
    FRAGMENTS
    + """
query GetEvaluationsForEvaluationIdQuery($evaluationId: ID!) {
    assetConditionEvaluationsForEvaluationId(evaluationId: $evaluationId) {
        ... on AssetConditionEvaluationRecords {
            records {
                evaluationId
                numRequested
                assetKey {
                    path
                }
                evaluation {
                    ...evaluationFields
                }
            }
        }
    }
}
"""
)

QUERY = (
    ENTITY_FRAGMENT
    + """
query GetEvaluationsQuery($assetKey: AssetKeyInput!, $limit: Int!, $cursor: String) {
    assetConditionEvaluationRecordsOrError(assetKey: $assetKey, limit: $limit, cursor: $cursor) {
        ... on AssetConditionEvaluationRecords {
            records {
                evaluationId
                isLegacy
                numRequested
                assetKey {
                    path
                }
                rootUniqueId
                evaluationNodes {
                    operatorType
                    userLabel
                    expandedLabel
                    startTimestamp
                    endTimestamp
                    numTrue
                    uniqueId
                    childUniqueIds
                    entityKey {
                        ...entityKeyFragment
                    }
                }
            }
        }
    }
}
"""
)

TRUE_PARTITIONS_QUERY = """
query GetTruePartitions($assetKey: AssetKeyInput!, $evaluationId: ID!, $nodeUniqueId: String!) {
    truePartitionsForAutomationConditionEvaluationNode(assetKey: $assetKey, evaluationId: $evaluationId, nodeUniqueId: $nodeUniqueId)
}
"""

QUERY_WITH_SINCE_METADATA = (
    ENTITY_FRAGMENT
    + """
query GetEvaluationsWithSinceMetadata($assetKey: AssetKeyInput!, $limit: Int!, $cursor: String) {
    assetConditionEvaluationRecordsOrError(assetKey: $assetKey, limit: $limit, cursor: $cursor) {
        ... on AssetConditionEvaluationRecords {
            records {
                evaluationId
                isLegacy
                numRequested
                assetKey {
                    path
                }
                rootUniqueId
                evaluationNodes {
                    operatorType
                    userLabel
                    expandedLabel
                    startTimestamp
                    endTimestamp
                    numTrue
                    uniqueId
                    childUniqueIds
                    entityKey {
                        ...entityKeyFragment
                    }
                    sinceMetadata {
                        triggerEvaluationId
                        triggerTimestamp
                        resetEvaluationId
                        resetTimestamp
                    }
                }
            }
        }
    }
}
"""
)


class TestAssetConditionEvaluations(ExecutingGraphQLContextTestMatrix):
    def test_auto_materialize_sensor(self, graphql_context: WorkspaceRequestContext):
        sensor_origin = RemoteInstigatorOrigin(
            repository_origin=infer_repository(graphql_context).get_remote_origin(),
            instigator_name="my_auto_materialize_sensor",
        )

        check.not_none(graphql_context.instance.schedule_storage).add_instigator_state(
            InstigatorState(
                sensor_origin,
                InstigatorType.SENSOR,
                status=InstigatorStatus.RUNNING,
                instigator_data=SensorInstigatorData(
                    sensor_type=SensorType.AUTO_MATERIALIZE,
                    cursor=asset_daemon_cursor_to_instigator_serialized_cursor(
                        AssetDaemonCursor.empty(12345)
                    ),
                ),
            )
        )

        with patch(
            graphql_context.instance.__class__.__module__
            + "."
            + graphql_context.instance.__class__.__name__
            + ".auto_materialize_use_sensors",
            new_callable=PropertyMock,
        ) as mock_my_property:
            mock_my_property.return_value = False

            results = execute_dagster_graphql(
                graphql_context,
                AUTO_MATERIALIZE_POLICY_SENSORS_QUERY,
                variables={
                    "assetKey": {"path": ["fresh_diamond_bottom"]},
                },
            )
            assert results.data["assetNodeOrError"]["currentAutoMaterializeEvaluationId"] == "0"

        with patch(
            graphql_context.instance.__class__.__module__
            + "."
            + graphql_context.instance.__class__.__name__
            + ".auto_materialize_use_sensors",
            new_callable=PropertyMock,
        ) as mock_my_property:
            mock_my_property.return_value = True
            results = execute_dagster_graphql(
                graphql_context,
                AUTO_MATERIALIZE_POLICY_SENSORS_QUERY,
                variables={
                    "assetKey": {"path": ["fresh_diamond_bottom"]},
                },
            )

            assert any(
                instigator["name"] == "my_auto_materialize_sensor"
                for instigator in results.data["assetNodeOrError"]["targetingInstigators"]
            )
            assert results.data["assetNodeOrError"]["currentAutoMaterializeEvaluationId"] == "12345"

    def _get_node(
        self, unique_id: str, evaluations: Sequence[Mapping[str, Any]]
    ) -> Mapping[str, Any]:
        return next(iter([node for node in evaluations if node["uniqueId"] == unique_id]))

    def _get_condition_evaluation(
        self,
        asset_key: AssetKey,
        description: str,
        partitions_def: PartitionsDefinition,
        true_partition_keys: Sequence[str],
        candidate_partition_keys: Sequence[str] | None = None,
        child_evaluations: Sequence[AutomationConditionEvaluation] | None = None,
    ) -> AutomationConditionEvaluation:
        return AutomationConditionEvaluation(
            condition_snapshot=AutomationConditionNodeSnapshot(
                class_name="...",
                description=description,
                unique_id=str(random.randint(0, 100000000)),
            ),
            true_subset=SerializableEntitySubset(
                key=asset_key,
                value=partitions_def.subset_with_partition_keys(true_partition_keys),
            ),
            candidate_subset=SerializableEntitySubset(
                key=asset_key,
                value=partitions_def.subset_with_partition_keys(candidate_partition_keys),
            )
            if candidate_partition_keys
            else HistoricalAllPartitionsSubsetSentinel(),
            start_timestamp=123,
            end_timestamp=456,
            child_evaluations=child_evaluations or [],
            subsets_with_metadata=[],
        )

    def test_get_evaluations_with_partitions(self, graphql_context: WorkspaceRequestContext):
        asset_key = AssetKey("upstream_static_partitioned_asset")
        partitions_def = StaticPartitionsDefinition(["a", "b", "c", "d", "e", "f"])
        results = execute_dagster_graphql(
            graphql_context,
            LEGACY_QUERY,
            variables={
                "assetKey": {"path": ["upstream_static_partitioned_asset"]},
                "limit": 10,
                "cursor": None,
            },
        )
        assert results.data["assetConditionEvaluationRecordsOrError"] == {"records": []}

        evaluation = self._get_condition_evaluation(
            asset_key,
            "All of",
            partitions_def,
            ["a", "b"],
            child_evaluations=[
                self._get_condition_evaluation(
                    asset_key,
                    "Any of",
                    partitions_def,
                    ["a", "b", "c"],
                    child_evaluations=[
                        self._get_condition_evaluation(
                            asset_key, "parent_updated", partitions_def, ["a", "c"]
                        ),
                        self._get_condition_evaluation(asset_key, "missing", partitions_def, ["b"]),
                        self._get_condition_evaluation(asset_key, "other", partitions_def, []),
                    ],
                ),
                self._get_condition_evaluation(
                    asset_key,
                    "Not",
                    partitions_def,
                    ["a", "b"],
                    candidate_partition_keys=["a", "b", "c"],
                    child_evaluations=[
                        self._get_condition_evaluation(
                            asset_key,
                            "Any of",
                            partitions_def,
                            ["c"],
                            ["a", "b", "c"],
                            child_evaluations=[
                                self._get_condition_evaluation(
                                    asset_key,
                                    "parent missing",
                                    partitions_def,
                                    ["c"],
                                    ["a", "b", "c"],
                                ),
                                self._get_condition_evaluation(
                                    asset_key,
                                    "parent outdated",
                                    partitions_def,
                                    [],
                                    ["a", "b", "c"],
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        )

        check.not_none(
            graphql_context.instance.schedule_storage
        ).add_auto_materialize_asset_evaluations(
            evaluation_id=10,
            asset_evaluations=[
                AutomationConditionEvaluationWithRunIds(
                    evaluation=evaluation, run_ids=frozenset({"runid1", "runid2"})
                )
            ],
        )

        results = execute_dagster_graphql(
            graphql_context,
            LEGACY_QUERY,
            variables={
                "assetKey": {"path": ["upstream_static_partitioned_asset"]},
                "limit": 10,
                "cursor": None,
            },
        )

        records = results.data["assetConditionEvaluationRecordsOrError"]["records"]
        assert len(records) == 1

        assert records[0]["numRequested"] == 2
        evaluation = records[0]["evaluation"]

        # all nodes in the tree
        assert len(evaluation["evaluationNodes"]) == 9

        rootNode = evaluation["evaluationNodes"][0]
        assert rootNode["uniqueId"] == evaluation["rootUniqueId"]
        assert rootNode["description"] == "All of"
        assert rootNode["numTrue"] == 2
        assert len(rootNode["childUniqueIds"]) == 2

        notNode = self._get_node(rootNode["childUniqueIds"][1], evaluation["evaluationNodes"])
        assert notNode["description"] == "Not"
        assert notNode["numTrue"] == 2

        skipNode = self._get_node(notNode["childUniqueIds"][0], evaluation["evaluationNodes"])
        assert skipNode["description"] == "Any of"
        assert skipNode["numTrue"] == 1

        evaluationId = records[0]["evaluationId"]
        uniqueId = skipNode["uniqueId"]
        results = execute_dagster_graphql(
            graphql_context,
            TRUE_PARTITIONS_QUERY,
            variables={
                "assetKey": {"path": ["upstream_static_partitioned_asset"]},
                "evaluationId": evaluationId,
                "nodeUniqueId": uniqueId,
            },
        )

        assert set(results.data["truePartitionsForAutomationConditionEvaluationNode"]) == {"c"}

        # test one of the true partitions
        specific_result = execute_dagster_graphql(
            graphql_context,
            LEGACY_QUERY_FOR_SPECIFIC_PARTITION,
            variables={
                "assetKey": {"path": ["upstream_static_partitioned_asset"]},
                "partition": "b",
                "evaluationId": 10,
            },
        )

        evaluation = specific_result.data["assetConditionEvaluationForPartition"]
        assert len(evaluation["evaluationNodes"]) == 9

        rootNode = evaluation["evaluationNodes"][0]
        assert rootNode["uniqueId"] == evaluation["rootUniqueId"]

        assert rootNode["description"] == "All of"
        assert rootNode["status"] == "TRUE"
        assert len(rootNode["childUniqueIds"]) == 2

        notNode = self._get_node(rootNode["childUniqueIds"][1], evaluation["evaluationNodes"])
        assert notNode["description"] == "Not"
        assert notNode["status"] == "TRUE"

        skipNode = self._get_node(notNode["childUniqueIds"][0], evaluation["evaluationNodes"])
        assert skipNode["description"] == "Any of"
        assert skipNode["status"] == "FALSE"

        # test one of the false partitions
        specific_result = execute_dagster_graphql(
            graphql_context,
            LEGACY_QUERY_FOR_SPECIFIC_PARTITION,
            variables={
                "assetKey": {"path": ["upstream_static_partitioned_asset"]},
                "partition": "d",
                "evaluationId": 10,
            },
        )

        evaluation = specific_result.data["assetConditionEvaluationForPartition"]
        assert len(evaluation["evaluationNodes"]) == 9

        rootNode = evaluation["evaluationNodes"][0]
        assert rootNode["uniqueId"] == evaluation["rootUniqueId"]

        assert rootNode["description"] == "All of"
        assert rootNode["status"] == "FALSE"
        assert len(rootNode["childUniqueIds"]) == 2

        notNode = self._get_node(rootNode["childUniqueIds"][1], evaluation["evaluationNodes"])
        assert notNode["description"] == "Not"
        assert notNode["status"] == "SKIPPED"

        skipNode = self._get_node(notNode["childUniqueIds"][0], evaluation["evaluationNodes"])
        assert skipNode["description"] == "Any of"
        assert skipNode["status"] == "SKIPPED"

    def test_get_evaluations_with_partitions_updated(
        self, graphql_context: WorkspaceRequestContext
    ):
        @asset(
            partitions_def=StaticPartitionsDefinition(["a", "b", "c", "d"]),
            automation_condition=AutomationCondition.eager().with_label("blah"),
            deps=["up"],
        )
        def A() -> None: ...

        results = execute_dagster_graphql(
            graphql_context,
            QUERY,
            variables={"assetKey": {"path": ["A"]}, "limit": 10, "cursor": None},
        )
        assert results.data == {"assetConditionEvaluationRecordsOrError": {"records": []}}

        result = evaluate_automation_conditions([A], DagsterInstance.ephemeral())

        check.not_none(
            graphql_context.instance.schedule_storage
        ).add_auto_materialize_asset_evaluations(
            evaluation_id=10,
            asset_evaluations=[
                AutomationConditionEvaluationWithRunIds(
                    evaluation=result.results[0].serializable_evaluation,
                    run_ids=frozenset({"runid1"}),
                )
            ],
        )

        results = execute_dagster_graphql(
            graphql_context,
            QUERY,
            variables={"assetKey": {"path": ["A"]}, "limit": 10, "cursor": None},
        )

        records = results.data["assetConditionEvaluationRecordsOrError"]["records"]
        assert len(records) == 1

        record = records[0]
        assert not record["isLegacy"]
        assert record["numRequested"] == 0

        # all nodes in the tree
        assert len(record["evaluationNodes"]) == 35

        rootNode = record["evaluationNodes"][0]
        assert rootNode["uniqueId"] == record["rootUniqueId"]
        assert rootNode["userLabel"] == "blah"
        assert rootNode["expandedLabel"] == [
            "(in_latest_time_window)",
            "AND",
            "(((newly_missing) OR (any_deps_updated)) SINCE (handled))",
            "AND",
            "(NOT (any_deps_missing))",
            "AND",
            "(NOT (any_deps_in_progress))",
            "AND",
            "(NOT (in_progress))",
        ]
        assert rootNode["numTrue"] == 0
        assert rootNode["operatorType"] == "and"
        assert len(rootNode["childUniqueIds"]) == 5

        def _get_node(id):
            return next(n for n in record["evaluationNodes"] if n["uniqueId"] == id)

        not_any_deps_missing_node = _get_node(rootNode["childUniqueIds"][2])
        any_deps_missing_node = _get_node(not_any_deps_missing_node["childUniqueIds"][0])
        up_node = _get_node(any_deps_missing_node["childUniqueIds"][0])
        assert up_node["expandedLabel"] == ["up", "((missing) AND (NOT (will_be_requested)))"]

        evaluationId = record["evaluationId"]
        uniqueId = rootNode["uniqueId"]
        results = execute_dagster_graphql(
            graphql_context,
            TRUE_PARTITIONS_QUERY,
            variables={
                "assetKey": {"path": ["A"]},
                "evaluationId": evaluationId,
                "nodeUniqueId": uniqueId,
            },
        )
        assert set(results.data["truePartitionsForAutomationConditionEvaluationNode"]) == set()

        childNode = record["evaluationNodes"][1]
        assert childNode["userLabel"] is None
        assert childNode["expandedLabel"] == ["in_latest_time_window"]
        assert childNode["numTrue"] == 4
        assert len(childNode["childUniqueIds"]) == 0

        evaluationId = record["evaluationId"]
        uniqueId = childNode["uniqueId"]
        results = execute_dagster_graphql(
            graphql_context,
            TRUE_PARTITIONS_QUERY,
            variables={
                "assetKey": {"path": ["A"]},
                "evaluationId": evaluationId,
                "nodeUniqueId": uniqueId,
            },
        )

        assert set(results.data["truePartitionsForAutomationConditionEvaluationNode"]) == {
            "a",
            "b",
            "c",
            "d",
        }

    def test_since_metadata_field(self, graphql_context: WorkspaceRequestContext):
        """Test that the sinceMetadata field is correctly populated for SinceCondition evaluations."""
        asset_key = AssetKey("test_asset")
        partitions_def = StaticPartitionsDefinition(["a", "b"])

        # Create an evaluation with SinceCondition metadata
        since_metadata = SinceConditionData(
            trigger_evaluation_id=5,
            trigger_timestamp=1234567890.0,
            reset_evaluation_id=3,
            reset_timestamp=1234567880.0,
        ).to_metadata()

        evaluation: AutomationConditionEvaluation[EntityKey] = AutomationConditionEvaluation(
            condition_snapshot=AutomationConditionNodeSnapshot(
                class_name="SinceCondition",
                description="since_test",
                unique_id="since_node_123",
            ),
            true_subset=SerializableEntitySubset(
                key=asset_key,
                value=partitions_def.subset_with_partition_keys(["a"]),
            ),
            candidate_subset=HistoricalAllPartitionsSubsetSentinel(),
            start_timestamp=123.0,
            end_timestamp=456.0,
            child_evaluations=[],
            subsets_with_metadata=[],
            metadata=since_metadata,
        )

        check.not_none(
            graphql_context.instance.schedule_storage
        ).add_auto_materialize_asset_evaluations(
            evaluation_id=200,
            asset_evaluations=[
                AutomationConditionEvaluationWithRunIds(evaluation=evaluation, run_ids=frozenset())
            ],
        )

        # Query with the sinceMetadata field
        results = execute_dagster_graphql(
            graphql_context,
            QUERY_WITH_SINCE_METADATA,
            variables={"assetKey": {"path": ["test_asset"]}, "limit": 10, "cursor": None},
        )

        records = results.data["assetConditionEvaluationRecordsOrError"]["records"]
        assert len(records) == 1

        # Find the SinceCondition node
        evaluation_nodes = records[0]["evaluationNodes"]
        assert len(evaluation_nodes) == 1
        since_node = evaluation_nodes[0]

        # Verify sinceMetadata is populated correctly
        assert since_node["sinceMetadata"] is not None
        since_metadata_result = since_node["sinceMetadata"]

        assert since_metadata_result["triggerEvaluationId"] == "5"
        assert since_metadata_result["triggerTimestamp"] == 1234567890.0
        assert since_metadata_result["resetEvaluationId"] == "3"
        assert since_metadata_result["resetTimestamp"] == 1234567880.0



import csv
from collections import OrderedDict

from dagster import (
    DependencyDefinition,
    In,
    JobDefinition,
    OpDefinition,
    dagster_type_loader,
    repository,
)
from dagster._core.definitions.graph_definition import GraphDefinition
from dagster._core.types.dagster_type import PythonObjectDagsterType
from dagster._core.workspace.context import WorkspaceRequestContext
from dagster._legacy import OutputDefinition
from dagster_graphql.schema.roots.mutation import execution_params_from_graphql
from dagster_graphql.test.utils import execute_dagster_graphql, infer_job_selector

from dagster_graphql_tests.graphql.production_query import PRODUCTION_QUERY


@dagster_type_loader(str)
def df_input_schema(_context, path):
    with open(path, encoding="utf8") as fd:
        return [OrderedDict(sorted(x.items(), key=lambda x: x[0])) for x in csv.DictReader(fd)]


PoorMansDataFrame = PythonObjectDagsterType(
    python_type=list,
    name="PoorMansDataFrame",
    loader=df_input_schema,
)


def test_enum_query(graphql_context: WorkspaceRequestContext):
    selector = infer_job_selector(graphql_context, "job_with_enum_config")

    ENUM_QUERY = """
    query EnumQuery($selector: PipelineSelector!) {
      runConfigSchemaOrError(selector: $selector) {
        ... on RunConfigSchema {
          allConfigTypes {
            __typename
            key
            ... on EnumConfigType {
              values
              {
                value
                description
              }
            }
          }
        }
      }
    }
    """

    result = execute_dagster_graphql(
        graphql_context,
        ENUM_QUERY,
        {
            "selector": selector,
        },
    )

    assert not result.errors
    assert result.data

    enum_type_data = None

    for td in result.data["runConfigSchemaOrError"]["allConfigTypes"]:
        if td["key"] == "TestEnum":
            enum_type_data = td
            break

    assert enum_type_data
    assert enum_type_data["key"] == "TestEnum"
    assert enum_type_data["values"] == [
        {"value": "ENUM_VALUE_ONE", "description": "An enum value."},
        {"value": "ENUM_VALUE_TWO", "description": "An enum value."},
        {"value": "ENUM_VALUE_THREE", "description": "An enum value."},
    ]


TYPE_RENDER_QUERY = """
fragment innerInfo on ConfigType {
  key
  recursiveConfigTypes {
    key
  }
  ... on CompositeConfigType {
    fields {
      name
      configType {
        key
      }
      isRequired
    }
  }
}

query TypeRenderQuery($selector: PipelineSelector!) {
  pipelineOrError(params: $selector) {
    __typename
    ... on Pipeline {
      name
      solids {
        name
        definition {
          ... on SolidDefinition {
            configField {
              configType {
                ...innerInfo
                recursiveConfigTypes {
                  ...innerInfo
                }
              }
            }
          }
        }
      }
    }
  }
}
"""


def test_type_rendering(graphql_context: WorkspaceRequestContext):
    selector = infer_job_selector(graphql_context, "more_complicated_nested_config")
    result = execute_dagster_graphql(graphql_context, TYPE_RENDER_QUERY, {"selector": selector})
    assert not result.errors
    assert result.data


def define_circular_dependency_job():
    return JobDefinition(
        graph_def=GraphDefinition(
            name="circular_dependency_job",
            node_defs=[
                OpDefinition(
                    name="csolid",
                    ins={"num": In("num", PoorMansDataFrame)},  # pyright: ignore[reportArgumentType]
                    outs={"result": OutputDefinition(PoorMansDataFrame)},  # pyright: ignore[reportArgumentType]
                    compute_fn=lambda *_args: None,
                )
            ],
            dependencies={"csolid": {"num": DependencyDefinition("csolid")}},
        )
    )


@repository  # pyright: ignore[reportArgumentType]
def the_test_repository():
    return {"jobs": {"circular_dependency_job": define_circular_dependency_job}}


def test_pipeline_or_error_by_name(graphql_context: WorkspaceRequestContext):
    selector = infer_job_selector(graphql_context, "csv_hello_world_two")
    result = execute_dagster_graphql(
        graphql_context,
        """
    query NamedPipelineQuery($selector: PipelineSelector!) {
        pipelineOrError(params: $selector) {
            ... on Pipeline {
                name
                nodeNames
            }
        }
    }""",
        {"selector": selector},
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineOrError"]["name"] == "csv_hello_world_two"
    assert result.data["pipelineOrError"]["nodeNames"] == ["sum_op"]


def test_pipeline_or_error_by_name_not_found(graphql_context: WorkspaceRequestContext):
    selector = infer_job_selector(graphql_context, "foobar")
    result = execute_dagster_graphql(
        graphql_context,
        """
        query NamedPipelineQuery($selector: PipelineSelector!) {
            pipelineOrError(params: $selector) {
                __typename
                ... on Pipeline {
                    name
                }
            }
        }""",
        {"selector": selector},
    )

    assert not result.errors
    assert result.data
    assert result.data["pipelineOrError"]["__typename"] == "PipelineNotFoundError"


def test_production_query(graphql_context: WorkspaceRequestContext):
    result = execute_dagster_graphql(graphql_context, PRODUCTION_QUERY)

    assert not result.errors
    assert result.data


def test_params_from_graphql():
    selector = {
        "repositoryName": "repo",
        "repositoryLocationName": "rl",
        "jobName": "job",
    }
    assert execution_params_from_graphql(
        {
            "selector": selector,
        }
    )
    assert execution_params_from_graphql(
        {
            "selector": selector,
            "runConfigData": "{}",
        }
    )
    assert execution_params_from_graphql(
        {
            "selector": selector,
            "runConfigData": "\n\n",
        }
    )
    assert execution_params_from_graphql(
        {
            "selector": selector,
            "runConfigData": "",
        }
    )


def test_repo_not_found(graphql_context: WorkspaceRequestContext):
    selector = {
        "repositoryLocationName": "junk",
        "repositoryName": "junk",
    }
    result = execute_dagster_graphql(
        graphql_context,
        """
        query repo($selector: RepositorySelector!) {
          repositoryOrError(repositorySelector: $selector) {
            __typename
          }
        }""",
        {"selector": selector},
    )
    assert not result.errors
    assert result.data["repositoryOrError"]["__typename"] == "RepositoryNotFoundError"

    result = execute_dagster_graphql(
        graphql_context,
        """
        query repos($selector: RepositorySelector!) {
          repositoriesOrError(repositorySelector: $selector) {
            __typename
          }
        }""",
        {"selector": selector},
    )
    assert not result.errors
    assert result.data["repositoriesOrError"]["__typename"] == "RepositoryNotFoundError"



from dagster._core.test_utils import SingleThreadPoolExecutor, wait_for_futures
from dagster._daemon import get_default_daemon_logger
from dagster._daemon.sensor import execute_sensor_iteration
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    infer_instigation_selector,
    infer_repository_selector,
)

from dagster_graphql_tests.graphql.graphql_context_test_suite import (
    NonLaunchableGraphQLContextTestMatrix,
)

INSTIGATION_QUERY = """
query JobQuery($instigationSelector: InstigationSelector! $id: String) {
  instigationStateOrError(instigationSelector: $instigationSelector id: $id) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on InstigationState {
        id
        nextTick {
            timestamp
        }
    }
  }
}
"""

INSTIGATION_STATES_QUERY = """
query InstigationStatesQuery($id: String!) {
  instigationStatesOrError(repositoryID: $id) {
    __typename
    ... on PythonError {
      message
      stack
    }
    ... on InstigationStates {
      results {
        name
        instigationType
      }
    }
  }
}
"""


def _create_sensor_tick(graphql_context):
    logger = get_default_daemon_logger("SensorDaemon")
    futures = {}
    list(
        execute_sensor_iteration(
            graphql_context.process_context,
            logger,
            threadpool_executor=SingleThreadPoolExecutor(),
            submit_threadpool_executor=None,
            sensor_tick_futures=futures,
        )
    )
    wait_for_futures(futures)


class TestNextTickRepository(NonLaunchableGraphQLContextTestMatrix):
    def test_schedule_next_tick(self, graphql_context) -> None:
        repository_selector = infer_repository_selector(graphql_context)
        remote_repository = graphql_context.get_code_location(
            repository_selector["repositoryLocationName"]
        ).get_repository(repository_selector["repositoryName"])

        schedule_name = "no_config_job_hourly_schedule"
        schedule = remote_repository.get_schedule(schedule_name)
        selector = infer_instigation_selector(graphql_context, schedule_name)

        # need to be running in order to generate a future tick
        graphql_context.instance.start_schedule(schedule)
        result = execute_dagster_graphql(
            graphql_context,
            INSTIGATION_QUERY,
            variables={
                "id": schedule.get_compound_id().to_string(),
                "instigationSelector": selector,
            },
        )

        assert result.data
        assert result.data["instigationStateOrError"]["__typename"] == "InstigationState"
        next_tick = result.data["instigationStateOrError"]["nextTick"]
        assert next_tick

        # ensure legacy selector based impl works until removal

        selector = infer_instigation_selector(graphql_context, schedule_name)
        result = execute_dagster_graphql(
            graphql_context, INSTIGATION_QUERY, variables={"instigationSelector": selector}
        )

        assert result.data
        assert result.data["instigationStateOrError"]["__typename"] == "InstigationState"
        next_tick = result.data["instigationStateOrError"]["nextTick"]
        assert next_tick

    def test_sensor_next_tick(self, graphql_context):
        repository_selector = infer_repository_selector(graphql_context)
        remote_repository = graphql_context.get_code_location(
            repository_selector["repositoryLocationName"]
        ).get_repository(repository_selector["repositoryName"])

        sensor_name = "always_no_config_sensor_with_tags_and_metadata"
        sensor = remote_repository.get_sensor(sensor_name)
        selector = infer_instigation_selector(graphql_context, sensor_name)

        result = execute_dagster_graphql(
            graphql_context,
            INSTIGATION_QUERY,
            variables={
                "id": sensor.get_compound_id().to_string(),
                "instigationSelector": selector,
            },
        )

        assert result.data
        assert result.data["instigationStateOrError"]["__typename"] == "InstigationState", (
            result.data["instigationStateOrError"]
        )

        # need to be running and create a sensor tick in the last 30 seconds in order to generate a
        # future tick
        graphql_context.instance.start_sensor(sensor)
        _create_sensor_tick(graphql_context)

        result = execute_dagster_graphql(
            graphql_context,
            INSTIGATION_QUERY,
            variables={
                "id": sensor.get_compound_id().to_string(),
                "instigationSelector": selector,
            },
        )

        assert result.data
        assert result.data["instigationStateOrError"]["__typename"] == "InstigationState", (
            result.data["instigationStateOrError"]
        )
        next_tick = result.data["instigationStateOrError"]["nextTick"]
        assert next_tick

        # ensure legacy selector based lookup continues to work until removed

        result = execute_dagster_graphql(
            graphql_context, INSTIGATION_QUERY, variables={"instigationSelector": selector}
        )

        assert result.data
        assert result.data["instigationStateOrError"]["__typename"] == "InstigationState", (
            result.data["instigationStateOrError"]
        )
        next_tick = result.data["instigationStateOrError"]["nextTick"]
        assert next_tick

    def test_instigation_states(self, graphql_context) -> None:
        repository_selector = infer_repository_selector(graphql_context)
        remote_repository = graphql_context.get_code_location(
            repository_selector["repositoryLocationName"]
        ).get_repository(repository_selector["repositoryName"])

        schedule_name = "no_config_job_hourly_schedule"
        schedule = remote_repository.get_schedule(schedule_name)
        graphql_context.instance.start_schedule(schedule)

        result = execute_dagster_graphql(
            graphql_context,
            INSTIGATION_STATES_QUERY,
            variables={"id": remote_repository.get_compound_id().to_string()},
        )

        assert result.data
        assert result.data["instigationStatesOrError"]["__typename"] == "InstigationStates", (
            result.data["instigationStatesOrError"]
        )

        results = result.data["instigationStatesOrError"]["results"]
        assert results == [{"name": schedule_name, "instigationType": "SCHEDULE"}]



import pytest
from dagster._core.utils import make_new_run_id
from dagster_graphql import DagsterGraphQLClientError

from dagster_graphql_tests.client_tests.conftest import MockClient, python_client_test_suite

RUN_IDS = [make_new_run_id(), make_new_run_id(), make_new_run_id()]


@python_client_test_suite
def test_successful_run_termination(mock_client: MockClient):
    expected_result = None
    response = {
        "terminateRuns": {
            "terminateRunResults": [
                {"__typename": "TerminateRunSuccess", "run": {"runId": run_id}}
                for run_id in RUN_IDS
            ]
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    actual_result = mock_client.python_client.terminate_runs(RUN_IDS)
    assert actual_result == expected_result


@python_client_test_suite
def test_complete_failure_run_not_found(mock_client: MockClient):
    error_messages = [("RunNotFoundError", f"Run Id {run_id} not found") for run_id in RUN_IDS]
    error_message_string = f"All run terminations failed: {error_messages}"
    response = {
        "terminateRuns": {
            "terminateRunResults": [
                {
                    "__typename": "RunNotFoundError",
                    "message": f"Run Id {run_id} not found",
                    "runId": run_id,
                }
                for run_id in RUN_IDS
            ]
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.terminate_runs(RUN_IDS)

    assert str(exc_info.value.args[0]) == "TerminateRunsError"
    assert exc_info.value.args[1] == error_message_string


@python_client_test_suite
def test_partial_failure_due_to_not_found_error(mock_client: MockClient):
    failed_ids = [RUN_IDS[0]]
    success_ids = RUN_IDS[1:]

    error_messages = [("RunNotFoundError", f"Run Id {run_id} not found") for run_id in failed_ids]
    error_message_string = f"Some runs could not be terminated: {error_messages}"

    response = {
        "terminateRuns": {
            "terminateRunResults": [
                {"__typename": "TerminateRunSuccess", "runId": run_id} for run_id in success_ids
            ]
            + [
                {
                    "__typename": "RunNotFoundError",
                    "message": f"Run Id {run_id} not found",
                    "runId": run_id,
                }
                for run_id in failed_ids
            ]
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.terminate_runs(RUN_IDS)

    assert str(exc_info.value.args[0]) == "TerminateRunsError"
    assert exc_info.value.args[1] == error_message_string


@python_client_test_suite
def test_complete_failure_generic_error(mock_client: MockClient):
    error_messages = [("TerminateRunFailure", "Unable to terminate run") for _ in RUN_IDS]
    error_message_string = f"All run terminations failed: {error_messages}"
    response = {
        "terminateRuns": {
            "terminateRunResults": [
                {"__typename": "TerminateRunFailure", "message": "Unable to terminate run"}
                for _ in RUN_IDS
            ]
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.terminate_runs(RUN_IDS)

    assert str(exc_info.value.args[0]) == "TerminateRunsError"
    assert exc_info.value.args[1] == error_message_string


@python_client_test_suite
def test_partial_failure_generic_error(mock_client: MockClient):
    success_ids = [RUN_IDS[0]]
    failed_ids = RUN_IDS[1:]

    error_messages = [("TerminateRunFailure", "Unable to terminate run") for _ in failed_ids]
    error_message_string = f"Some runs could not be terminated: {error_messages}"

    response = {
        "terminateRuns": {
            "terminateRunResults": [
                {"__typename": "TerminateRunSuccess", "run": {"runId": run_id}}
                for run_id in success_ids
            ]
            + [
                {"__typename": "TerminateRunFailure", "message": "Unable to terminate run"}
                for run_id in failed_ids
            ]
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.terminate_runs(RUN_IDS)

    assert str(exc_info.value.args[0]) == "TerminateRunsError"
    assert exc_info.value.args[1] == error_message_string


@python_client_test_suite
def test_complete_failure_python_error(mock_client: MockClient):
    error_messages = [("PythonError", "Unable to terminate run") for _ in RUN_IDS]
    error_message_string = f"All run terminations failed: {error_messages}"
    response = {
        "terminateRuns": {
            "terminateRunResults": [
                {"__typename": "PythonError", "message": "Unable to terminate run"} for _ in RUN_IDS
            ]
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.terminate_runs(RUN_IDS)

    assert str(exc_info.value.args[0]) == "TerminateRunsError"
    assert exc_info.value.args[1] == error_message_string


@python_client_test_suite
def test_partial_failure_python_error(mock_client: MockClient):
    success_ids = [RUN_IDS[0]]
    failed_ids = RUN_IDS[1:]

    error_messages = [("PythonError", "Unable to terminate run") for _ in failed_ids]
    error_message_string = f"Some runs could not be terminated: {error_messages}"

    response = {
        "terminateRuns": {
            "terminateRunResults": [
                {"__typename": "TerminateRunSuccess", "run": {"runId": run_id}}
                for run_id in success_ids
            ]
            + [
                {"__typename": "PythonError", "message": "Unable to terminate run"}
                for _ in failed_ids
            ]
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.terminate_runs(RUN_IDS)

    assert str(exc_info.value.args[0]) == "TerminateRunsError"
    assert exc_info.value.args[1] == error_message_string


@python_client_test_suite
def test_terminate_runs_mixed_results(mock_client: MockClient):
    success_id = RUN_IDS[0]
    not_found_id = RUN_IDS[1]

    error_messages = [
        ("RunNotFoundError", f"Run Id {not_found_id} not found"),
        ("PythonError", "Internal server error occurred"),
    ]
    error_message_string = f"Some runs could not be terminated: {error_messages}"

    response = {
        "terminateRuns": {
            "terminateRunResults": [
                {"__typename": "TerminateRunSuccess", "run": {"runId": success_id}},
                {
                    "__typename": "RunNotFoundError",
                    "message": f"Run Id {not_found_id} not found",
                    "runId": not_found_id,
                },
                {
                    "__typename": "PythonError",
                    "message": "Internal server error occurred",
                    "stack": "Traceback info",
                },
            ]
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.terminate_runs(RUN_IDS)

    assert str(exc_info.value.args[0]) == "TerminateRunsError"
    assert exc_info.value.args[1] == error_message_string
